from .HigherOrderMeshingLine import HighOrderMeshLine
from .HigherOrderMeshingTri import HighOrderMeshTri_SEMISTABLE
from .HigherOrderMeshingTet import HighOrderMeshTet_SEMISTABLE
from .HigherOrderMeshingQuad import HighOrderMeshQuad
from .HigherOrderMeshingHex import HighOrderMeshHex


import numpy as np
from copy import deepcopy
from warnings import warn
import gc
from time import time
import multiprocessing as MP

from .GetInteriorCoordinates import GetInteriorNodesCoordinates
from Florence.Tensor import itemfreq, makezero, unique2d, remove_duplicates_2D
import Florence.ParallelProcessing.parmap as parmap

#--------------------------------------------------------------------------------------------------------------------------#
# SUPPLEMENTARY FUNCTIONS
def ElementLoopTet(elem,elements,points,MeshType,eps,Neval):
    xycoord_higher = GetInteriorNodesCoordinates(points[elements[elem,:],:],MeshType,elem,eps,Neval)
    return xycoord_higher


def HighOrderMeshTet_SEMISTABLE(C, mesh, Decimals=10, equally_spaced=False, check_duplicates=True,
    Zerofy=True, Parallel=False, nCPU=1, ComputeAll=True):

    from Florence.FunctionSpace import Tet
    from Florence.QuadratureRules.FeketePointsTet import FeketePointsTet
    from Florence.MeshGeneration.NodeArrangement import NodeArrangementTet

    # SWITCH OFF MULTI-PROCESSING FOR SMALLER PROBLEMS WITHOUT GIVING A MESSAGE
    if (mesh.elements.shape[0] < 500) and (C < 5):
        Parallel = False
        nCPU = 1

    if not equally_spaced:
        eps = FeketePointsTet(C)
        # COMPUTE BASES FUNCTIONS AT ALL NODAL POINTS
        Neval = np.zeros((4,eps.shape[0]),dtype=np.float64)
        hpBases = Tet.hpNodal.hpBases
        for i in range(4,eps.shape[0]):
            Neval[:,i] = hpBases(0,eps[i,0],eps[i,1],eps[i,2],Transform=1,EvalOpt=1)[0]
    else:
        from Florence.QuadratureRules.EquallySpacedPoints import EquallySpacedPointsTet
        eps =  EquallySpacedPointsTet(C)
        # COMPUTE BASES FUNCTIONS AT ALL NODAL POINTS
        hpBases = Tet.hpNodal.hpBases
        Neval = np.zeros((4,eps.shape[0]),dtype=np.float64)
        for i in range(4,eps.shape[0]):
            Neval[:,i]  = hpBases(0,eps[i,0],eps[i,1],eps[i,2],Transform=1,EvalOpt=1,equally_spaced=True)[0]

    # THIS IS NECESSARY FOR REMOVING DUPLICATES
    makezero(Neval, tol=1e-12)

    nodeperelem = mesh.elements.shape[1]
    renodeperelem = int((C+2.)*(C+3.)*(C+4.)/6.)
    left_over_nodes = renodeperelem - nodeperelem

    reelements = -1*np.ones((mesh.elements.shape[0],renodeperelem),dtype=np.int64)
    reelements[:,:4] = mesh.elements
    # TOTAL NUMBER OF (INTERIOR+EDGE+FACE) NODES
    iesize = np.int64(C*(C-1)*(C-2)/6. + 6.*C + 2*C*(C-1))
    repoints = np.zeros((mesh.points.shape[0]+iesize*mesh.elements.shape[0],3),dtype=np.float64)
    repoints[:mesh.points.shape[0],:]=mesh.points

    telements = time()

    xycoord_higher=[]; ParallelTuple1=[]
    if Parallel:
        # GET HIGHER ORDER COORDINATES - PARALLEL
        ParallelTuple1 = parmap.map(ElementLoopTet,np.arange(0,mesh.elements.shape[0]),mesh.elements,mesh.points,'tet',eps,
            Neval,pool=MP.Pool(processes=nCPU))

    maxNode = np.max(reelements)
    for elem in range(0,mesh.elements.shape[0]):
        # maxNode = np.max(reelements) # BIG BOTTLENECK
        if Parallel:
            xycoord_higher = ParallelTuple1[elem]
        else:
            xycoord =  mesh.points[mesh.elements[elem,:],:]
            # GET HIGHER ORDER COORDINATES
            xycoord_higher = GetInteriorNodesCoordinates(xycoord,'tet',elem,eps,Neval)

        # EXPAND THE ELEMENT CONNECTIVITY
        newElements = np.arange(maxNode+1,maxNode+1+left_over_nodes)
        reelements[elem,4:] = newElements
        # INSTEAD COMPUTE maxNode BY INDEXING
        maxNode = newElements[-1]

        repoints[mesh.points.shape[0]+elem*iesize:mesh.points.shape[0]+(elem+1)*iesize] = xycoord_higher[4:,:]

    if Parallel:
        del ParallelTuple1

    telements = time()-telements
    #--------------------------------------------------------------------------------------
    # NOW REMOVE DUPLICATED POINTS
    tnodes = time()

    nnode_linear = mesh.points.shape[0]
    # KEEP ZEROFY ON, OTHERWISE YOU GET STRANGE BEHVAIOUR
    # rounded_repoints = makezero(repoints[nnode_linear:,:].copy())
    rounded_repoints = repoints[nnode_linear:,:].copy()
    makezero(rounded_repoints)
    rounded_repoints = np.round(rounded_repoints,decimals=Decimals)

    _, idx_repoints, inv_repoints = unique2d(rounded_repoints,order=False,
        consider_sort=False,return_index=True,return_inverse=True)
    # idx_repoints.sort()
    del rounded_repoints

    idx_repoints = np.concatenate((np.arange(nnode_linear),idx_repoints+nnode_linear))
    repoints = repoints[idx_repoints,:]

    unique_reelements, inv_reelements = np.unique(reelements[:,4:],return_inverse=True)
    unique_reelements = unique_reelements[inv_repoints]
    reelements = unique_reelements[inv_reelements]
    reelements = reelements.reshape(mesh.elements.shape[0],renodeperelem-4)
    reelements = np.concatenate((mesh.elements,reelements),axis=1)

    # SANITY CHECK fOR DUPLICATES
    #---------------------------------------------------------------------#
    # NOTE THAT THIS REMAPS THE ELEMENT CONNECTIVITY FOR THE WHOLE MESH
    # AND AS A RESULT THE FIRST FEW COLUMNS WOULD NO LONGER CORRESPOND TO
    # LINEAR CONNECTIVITY
    if check_duplicates:
        last_shape = repoints.shape[0]
        deci = int(Decimals)-2
        if Decimals < 6:
            deci = Decimals
        repoints, idx_repoints, inv_repoints = remove_duplicates_2D(repoints, decimals=deci)
        unique_reelements, inv_reelements = np.unique(reelements,return_inverse=True)
        unique_reelements = unique_reelements[inv_repoints]
        reelements = unique_reelements[inv_reelements]
        reelements = reelements.reshape(mesh.elements.shape[0],renodeperelem)
        if last_shape != repoints.shape[0]:
            warn('Duplicated points generated in high order mesh. Lower the "Decimals". I have fixed it for now')
    #---------------------------------------------------------------------#

    tnodes = time() - tnodes
    #------------------------------------------------------------------------------------------


    # GET MESH EDGES AND FACES
    reedges = np.zeros((mesh.edges.shape[0],C+2))
    fsize = int((C+2.)*(C+3.)/2.)
    refaces = np.zeros((mesh.faces.shape[0],fsize),dtype=mesh.faces.dtype)

    # ComputeAll = False
    if ComputeAll == True:
        #------------------------------------------------------------------------------------------
        # BUILD FACES NOW
        tfaces = time()

        refaces = np.zeros((mesh.faces.shape[0],fsize))
        # DO NOT CHANGE THE FACES, BY RECOMPUTING THEM, AS THE LINEAR FACES CAN COME FROM
        # AN EXTERNAL MESH GENERATOR, WHOSE ORDERING MAY NOT BE THE SAME, SO JUST FIND WHICH
        # ELEMENTS CONTAIN THESE FACES
        face_to_elements = mesh.GetElementsWithBoundaryFacesTet()
        node_arranger = NodeArrangementTet(C)[0]

        refaces = reelements[face_to_elements[:,0][:,None],node_arranger[face_to_elements[:,1],:]].astype(mesh.faces.dtype)

        tfaces = time()-tfaces
        #------------------------------------------------------------------------------------------

        #------------------------------------------------------------------------------------------
        # BUILD EDGES NOW
        tedges = time()

        # BUILD A 2D MESH
        from Florence import Mesh
        tmesh = Mesh()
        tmesh.element_type = "tri"
        tmesh.elements = refaces
        tmesh.nelem = tmesh.elements.shape[0]
        # GET BOUNDARY EDGES
        reedges = tmesh.GetEdgesTri()
        del tmesh

        tedges = time()-tedges
        #------------------------------------------------------------------------------------------

    class nmesh(object):
        # """Construct pMesh"""
        points = repoints
        elements = reelements
        edges = np.array([[],[]])
        faces = np.array([[],[]])
        nnode = repoints.shape[0]
        nelem = reelements.shape[0]
        info = 'tet'
    if ComputeAll is True:
        nmesh.edges = reedges
        nmesh.faces = refaces

    gc.collect()


    # print '\nHigh order meshing timing:\n\t\tElement loop:\t '+str(telements)+' seconds\n\t\tNode loop:\t\t '+str(tnodes)+\
    #  ' seconds'+'\n\t\tEdge loop:\t\t '+str(tedges)+' seconds'+\
    #  '\n\t\tFace loop:\t\t '+str(tfaces)+' seconds\n'

    return nmesh


import numpy as np
from numpy import einsum
from Florence.Tensor import trace, Voigt
from .MaterialBase import Material
from Florence.LegendreTransform import LegendreTransform


class IdealDielectric(Material):
    """Ideal dielectric material model for Laplace problems
    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(IdealDielectric, self).__init__(mtype, ndim, **kwargs)
        # REQUIRES SEPARATELY
        self.nvar = 1
        self.energy_type = "internal_energy"
        self.legendre_transform = LegendreTransform()
        self.nature = "linear"
        self.fields = "electrostatics"

        if not hasattr(self,'eps_1'):
            if hasattr(self,'eps'):
                self.eps_1 = self.eps

        self.H_VoigtSize = self.ndim

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = True
        # self.has_low_level_dispatcher = False

    def KineticMeasures(self,F,ElectricFieldx, elem=0):
        # from Florence.MaterialLibrary.LLDispatch._IdealDielectric_ import KineticMeasures
        # D, H_Voigt = KineticMeasures(self,np.ascontiguousarray(F), ElectricFieldx)
        # return D, None, H_Voigt
        ndim = self.ndim
        eps_1 = float(self.eps_1)
        J = np.linalg.det(F)
        I = np.eye(ndim,ndim)
        D = einsum("i,ij->ij",eps_1/J,ElectricFieldx)[...,None]
        # H_Voigt = np.broadcast_to(1./eps_1*I,(J.shape[0],ndim,ndim))
        # Negative definite inverse is needed due to Legendre transform
        H_Voigt = np.einsum("i,jk",-eps_1/J,I)

        return np.ascontiguousarray(D), None, np.ascontiguousarray(H_Voigt)

    def Hessian(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        eps_1 = self.eps_1
        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        # self.dielectric_tensor = 1./eps_1*I
        # Negative definite inverse is needed due to Legendre transform
        self.dielectric_tensor = -eps_1/J*I
        return self.dielectric_tensor

    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        pass

    def ElectricDisplacementx(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        # D = self.legendre_transform.GetElectricDisplacement(self, StrainTensors, ElectricFieldx, elem, gcounter)
        eps_1 = self.eps_1
        J = StrainTensors['J'][gcounter]
        E = ElectricFieldx.reshape(self.ndim,1)
        D_exact = eps_1/J*E
        return D_exact

    def Permittivity(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        eps_1 = self.eps_1
        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        # self.dielectric_tensor = 1./eps_1*I
        # Negative definite inverse is needed due to Legendre transform
        self.dielectric_tensor = -eps_1/J*I
        return self.dielectric_tensor

    def ElectrostaticMeasures(self,F,ElectricFieldx, elem=0):
        return self.KineticMeasures(F,ElectricFieldx, elem)


#!/usr/bin/env python
from mpi4py import MPI
import numpy as np
from time import time


def runParallel():

    from Florence.FiniteElements.Assembly._LowLevelAssembly_ import _LowLevelAssemblyExplicit_Par_
    from Florence import DisplacementFormulation, DisplacementPotentialFormulation

    comm = MPI.Comm.Get_parent()
    size = comm.Get_size()
    rank = comm.Get_rank()

    T_all_size = np.empty(3,'i')
    comm.Bcast(T_all_size, root=0)
    funcs = None
    funcs = comm.bcast(funcs, root=0)

    nnode = T_all_size[0]
    ndim  = T_all_size[1]
    nvar  = T_all_size[2]
    T_all = np.zeros((nnode,nvar),np.float64)

    Eulerx = np.zeros((nnode,ndim),np.float64)
    comm.Bcast([Eulerx, MPI.DOUBLE], root=0)

    Eulerp = np.zeros((nnode),np.float64)
    comm.Bcast([Eulerp, MPI.DOUBLE], root=0)

    for proc in range(size):
        if proc == rank:
            functor = funcs[proc]
            pnodes = funcs[proc].pnodes
            # tt = time()
            T = _LowLevelAssemblyExplicit_Par_(functor.formulation.function_spaces[0],
                functor.formulation, functor.mesh, functor.material, Eulerx[pnodes,:], Eulerp[pnodes])
            T_all[pnodes,:] += T.reshape(pnodes.shape[0],nvar)
            # print(time()-tt)


    comm.Reduce([T_all, MPI.DOUBLE], None, root=0)
    comm.Disconnect()



if __name__ == "__main__":
    runParallel()


from numpy import einsum, asarray, eye
from .MaterialBase import Material
from Florence.Tensor import trace, Voigt


#####################################################################################################
                                    # NEARLY INCOMPRESSIBLE NEOHOOKEAN 
#####################################################################################################


class NearlyIncompressibleNeoHookean(Material):
    """Material model for nearly incompressible neo-Hookean with the following internal energy:

        W(C) = mu/2*J**(-2/3)*(C:I)     # for isochoric part
        U(J) = k/2*(J-1)**2             # for volumetric part

        """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(NearlyIncompressibleNeoHookean, self).__init__(mtype, ndim, **kwargs)

        self.is_nearly_incompressible = True
        self.is_compressible = False

        self.is_transversely_isotropic = False
        self.energy_type = "internal_energy"
        self.nature = "nonlinear"
        self.fields = "mechanics"

        if self.ndim==3:
            self.H_VoigtSize = 6
        elif self.ndim==2:
            self.H_VoigtSize = 3

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = False
        

    def Hessian(self, StrainTensors, ElectricFieldx=0, elem=0, gcounter=0):
        """Hessian split into isochoroic and volumetric parts"""
        
        I = StrainTensors['I']
        b = StrainTensors['b'][gcounter]
        J = StrainTensors['J'][gcounter]
        mu = self.mu

        # ISOCHORIC
        H_Voigt = 2*mu*J**(-5./3.)*(1./9.*trace(b)*einsum('ij,kl',I,I) - \
            1./3.*(einsum('ij,kl',b,I) + einsum('ij,kl',I,b)) +\
            1./6.*trace(b)*(einsum('ik,jl',I,I) + einsum('il,jk',I,I)) )
        # VOLUMETRIC
        H_Voigt += self.pressure[elem]*(einsum('ij,kl',I,I) - (einsum('ik,jl',I,I) + einsum('il,jk',I,I))) 

        H_Voigt = Voigt(H_Voigt,1)

        self.H_VoigtSize = H_Voigt.shape[0]

        return H_Voigt


    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]

        mu = self.mu
        stress = mu*J**(-5./3.)*(b - 1./3.*trace(b)*I) 
        stress += self.pressure[elem]*I
            
        return stress












########################################################
# import numpy as np
# from Florence.Tensor import trace
# #####################################################################################################
#                                 # NEARLY INCOMPRESSIBLE NEOHOOKEAN
#                                 # W = mu/2*C:I + k/2*(J-1)**2                               
# #####################################################################################################


# class NearlyIncompressibleNeoHookean(object):
#     """ A nearly incompressible neo-Hookean material model whose energy functional is given by:

#                 W = mu/2*C:I + k/2*(J-1)**2

#             This is an incorrect internal energy for incompressibility as C:I is not pure 
#             deviatoric. It is missing a factor J^{-2/3}


#         """

#     def __init__(self, ndim):
#         super(NearlyIncompressibleNeoHookean, self).__init__()
#         self.ndim = ndim
#         self.nvar = self.ndim

#     def Hessian(self,MaterialArgs,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):

#         # Using Einstein summation (using numpy einsum call)
#         d = np.einsum

#         # Get material constants (5 in this case)
#         mu = MaterialArgs.mu
#         lamb = MaterialArgs.lamb

#         I = StrainTensors['I']
#         J = StrainTensors['J'][gcounter]
#         b = StrainTensors['b'][gcounter]

#         # Update Lame constants
#         kappa = lamb+2.0*mu/3.0


#         H_Voigt = Voigt( kappa*(2.0*J-1)*d('ij,kl',I,I)-kappa*(J-1)*(d('ik,jl',I,I)+d('il,jk',I,I)) ,1)
        
#         MaterialArgs.H_VoigtSize = H_Voigt.shape[0]

#         return H_Voigt



#     def CauchyStress(self,MaterialArgs,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

#         I = StrainTensors['I']
#         J = StrainTensors['J'][gcounter]
#         b = StrainTensors['b'][gcounter]

#         mu = MaterialArgs.mu
#         lamb = MaterialArgs.lamb
#         kappa = lamb+2.0*mu/3.0

#         return 1.0*mu/J*b+(kappa*(J-1.0))*I 


#     def ElectricDisplacementx(self,MaterialArgs,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
#         ndim = StrainTensors['I'].shape[0]
#         return np.zeros((ndim,1))



import numpy as np

def KinematicMeasures(F,analysis_nature):
    """Computes the kinematic measures at all Gauss points for a given element.

    input:

        F:                  [ndarray (nofGauss x ndim x ndim)] Deformation gradient tensor evaluated at all integration points
        analysis_nature:       [str] type of analysis i.e. linear or non-linear 


    returns:

        StrainTensors:      [dict] a dictionary containing kinematic measures such as F, J, b etc evaluated at all integration points   
    """
    
    assert F.ndim == 3


    StrainTensors = {'F':F, 'J':np.linalg.det(F), 'b':np.einsum('ijk,ilk->ijl',F,F),
    'I':np.eye(F.shape[1],F.shape[1],dtype=np.float64)}


    # if analysis_nature == 'Nonlinear':
    #   # ADDITIONAL POLYCONVEX MEASURES - ACTIVATE IF NECESSARY
    #   # StrainTensors['H'] = np.einsum('i,ijk->ijk',StrainTensors['J'],np.einsum('ikj',np.linalg.inv(StrainTensors['F'])))
    #   # StrainTensors['H'] = np.einsum('i,ikj->ijk',StrainTensors['J'],np.linalg.inv(StrainTensors['F'])) # or
    #   # StrainTensors['g'] = np.einsum('ijk,ilk->ijl',StrainTensors['H'],StrainTensors['H'])
    #   # StrainTensors['C'] = np.einsum('ikj,ikl->ijl',StrainTensors['F'],StrainTensors['F']) 
    #   # StrainTensors['G'] = np.einsum('ikj,ikl->ijl',StrainTensors['H'],StrainTensors['H'])
    #   # StrainTensors['detC'] = np.einsum('i,i->i',StrainTensors['J'],StrainTensors['J'])
    #   pass


    # if analysis_nature=='linear':
    #     # LINEARISED KINEMATICS
    #     # MATERIAL GRADIENT OF DISPLACEMENT
    #     StrainTensors['Gradu'] = F - StrainTensors['I']
    #     # SMALL STRAIN TENSOR IS THE LINEARISED VERSION OF GREEN-LAGRANGE STRAIN TENSOR
    #     StrainTensors['strain'] = 0.5*(StrainTensors['Gradu'] + np.einsum('ikj',StrainTensors['Gradu']))


    # LINEARISED KINEMATICS
    # MATERIAL GRADIENT OF DISPLACEMENT
    StrainTensors['Gradu'] = F - StrainTensors['I']
    # SMALL STRAIN TENSOR IS THE LINEARISED VERSION OF GREEN-LAGRANGE STRAIN TENSOR
    StrainTensors['strain'] = 0.5*(StrainTensors['Gradu'] + np.einsum('ikj',StrainTensors['Gradu']))

        
    return StrainTensors









def KinematicMeasures_NonVectorised(F,AnalysisType,ncounter):
    """Computes the kinematic measures at all Gauss points for a given element.

    input:

        F:                  [ndarray (nofGauss x ndim x ndim)] Deformation gradient tensor evaluated at all integration points
        AnalysisType:       [str] type of analysis i.e. linear or non-linear 
        ncounter:           [int] no of Gauss points


    returns:

        StrainTensors:      [dict] a dictionary containing kinematic measures such as J, b etc evaluated at all integration points  
    """
    
    assert len(F.shape) == 2


    StrainTensors = {'F':[F]*ncounter, 'J':[np.linalg.det(F)]*ncounter, 'b':[np.dot(F,F.T)]*ncounter,
    'I':np.eye(F.shape[1],F.shape[1],dtype=np.float64)}


    if AnalysisType == 'Nonlinear':
        # ADDITIONAL POLYCONVEX MEASURES - ACTIVATE IF NECESSARY
        # StrainTensors['H'] = [StrainTensors['J'][0]*np.linalg.inv(F)]*ncounter
        pass


    elif AnalysisType=='Linear':
        # LINEARISED KINEMATICS
        # MATERIAL GRADIENT OF DISPLACEMENT
        StrainTensors['Gradu'] = F - StrainTensors['I']
        # SMALL STRAIN TENSOR IS THE LINEARISED VERSION OF GREEN-LAGRANGE STRAIN TENSOR
        StrainTensors['strain'] = [0.5*(StrainTensors['Gradu'] + StrainTensors['Gradu'].T)]*ncounter
        # RE-ASSIGN
        StrainTensors['Gradu'] = [F - StrainTensors['I']]*ncounter


    return StrainTensors






#                   OLD VERSION - KEPT FOR LEGACY/DEBUG
#--------------------------------------------------------------------------------#
# class KinematicMeasures(object):
#   """docstring for KinematicMeasures"""
#   def __init__(self, F):
#       super(KinematicMeasures, self).__init__()
#       self.F = F

#   def Compute(self,AnalysisType):

#       # F = self.F
#       self.J = np.linalg.det(self.F)
#       self.I = np.eye(self.F.shape[0],self.F.shape[0],dtype=np.float64)

#       self.b = np.dot(self.F,self.F.T)
#       # self.C = np.dot(self.F.T,self.F)

#       # self.Gradu = self.F - self.I
#       # self.strain = 0.5*(self.Gradu + self.Gradu.T)

#       if AnalysisType=='Nonlinear':
#           # self.C = np.dot(self.F.T,self.F)
#           # self.b = np.dot(self.F,self.F.T)
#           # self.E = 0.5*(self.C-self.I)
#           # self.e = 0.5*(self.I-np.linalg.inv(self.b)) 

#           # self.Gradu = self.F - self.I
#           # self.strain = 0.5*(self.Gradu + self.Gradu.T)
            

#           # POLYCONVEX KINEMATICS MEASURE
#           self.H = self.J*np.linalg.inv(self.F).T
#           # self.G = self.H.T*self.H
#       elif AnalysisType=='Linear':
#           # LINEARISED KINEMATICS
#           # MATERIAL GRADIENT OF DISPLACEMENT
#           self.Gradu = self.F - self.I
#           # SMALL STRAIN TENSOR IS THE LINEARISED VERSION OF GREEN-LAGRANGE STRAIN TENSOR
#           self.strain = 0.5*(self.Gradu + self.Gradu.T)

#       return self





from __future__ import division
import numpy as np
from numpy import einsum
from .MaterialBase import Material
from Florence.Tensor import trace, Voigt
from math import sqrt


#####################################################################################################
                                # NEARLY INCOMPRESSIBLE MOONEY-RIVLIN
#####################################################################################################


class NearlyIncompressibleMooneyRivlin(Material):
    """ A nearly incompressible Mooney-Rivlin material model whose energy functional is given by:
        
            W(C,G,J**2) = alpha*J**(-2/3)*(C:I) + beta*J**(-2)*(G:I)**(3/2) + kappa/2*(J-1)**2

        Note that this energy is decomposed into deviatoric and volumetric components such that
        C:I and (G:I)**(3/2) contain only deviatoric contribution and the volumetric contribution
        is taken care of by the bulk modulus (kappa) term (J-1)**2 

        """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(NearlyIncompressibleMooneyRivlin, self).__init__(mtype, ndim, **kwargs)

        self.gamma=1.
        self.alpha = self.gamma*self.mu/2.
        self.beta = (self.mu - 2.*self.alpha)/3./sqrt(3.)
        # self.kappa = self.lamb+2.0*self.mu/3.0 # or
        self.kappa = self.lamb+4.0/3.0*self.alpha+2.0*sqrt(3.0)*self.beta

        self.is_transversely_isotropic = False
        self.energy_type = "internal_energy"
        self.nature = "nonlinear"
        self.fields = "mechanics"  

        if self.ndim==3:
            self.H_VoigtSize = 6
        elif self.ndim==2:
            self.H_VoigtSize = 3

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = True
        # self.has_low_level_dispatcher = False

    def KineticMeasures(self,F,ElectricFieldx=0, elem=0):
        from Florence.MaterialLibrary.LLDispatch._NearlyIncompressibleMooneyRivlin_ import KineticMeasures
        return KineticMeasures(self,np.ascontiguousarray(F))

        

    def Hessian(self,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):

        alpha = self.alpha
        beta = self.beta
        kappa = self.kappa
        

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        F = StrainTensors['F'][gcounter]
        # b=np.dot(F,F.T)
        H = J*np.linalg.inv(F).T
        g = np.dot(H,H.T)

        


        if self.ndim == 2:
            trb = trace(b)+1
            trg = trace(g)+J**2
        elif self.ndim == 3:
            trb = trace(b)
            trg = trace(g)


        H_Voigt = -4/3.*alpha*J**(-5/3.)*( einsum('ij,kl',b,I) + einsum('ij,kl',I,b) ) + \
                  4.*alpha/9.*J**(-5/3.)*trb*einsum('ij,kl',I,I) + \
                  2/3.*alpha*J**(-5/3.)*trb*( einsum('il,jk',I,I) + einsum('ik,jl',I,I) ) + \
          beta*J**(-3)*trg**(3./2.)* ( einsum('ij,kl',I,I) - einsum('ik,jl',I,I) - einsum('il,jk',I,I) ) - \
          3.*beta*J**(-3)*trg**(1./2.)*( einsum('ij,kl',I,g) + einsum('ij,kl',g,I) ) + \
          6.*beta*J**(-3)*trg**(1./2.)*( einsum('ik,jl',I,g) + einsum('il,jk',g,I) ) + \
          3.*beta*J**(-3)*trg**(-1./2.)*( einsum('ij,kl',g,g) )   + \
          kappa*(2.0*J-1)*einsum('ij,kl',I,I) - kappa*(J-1)*(einsum('ik,jl',I,I)+einsum('il,jk',I,I))         # #


        # # WITH PRE-COMPUTED IDENTITY TENSORS
        # H_Voigt = -4/3.*alpha*J**(-5/3.)*( einsum('ij,kl',b,I) + einsum('ij,kl',I,b) ) + \
        #             4.*alpha/9.*J**(-5/3.)*trb*self.Iijkl + \
        #             2/3.*alpha*J**(-5/3.)*trb*self.Iikjl + \
        #     beta*J**(-3)*trg**(3./2.)*( self.Iijkl - self.Iikjl ) - \
        #     3.*beta*J**(-3)*trg**(1./2.)*( einsum('ij,kl',I,g) + einsum('ij,kl',g,I) ) + \
        #     6.*beta*J**(-3)*trg**(1./2.)*( einsum('ik,jl',I,g) + einsum('il,jk',g,I) ) + \
        #     3.*beta*J**(-3)*trg**(-1./2.)*( einsum('ij,kl',g,g) )   + \
        #     kappa*(2.0*J-1)*self.Iijkl - kappa*(J-1)*self.Iikjl 


        H_Voigt = Voigt( H_Voigt ,1)
        
        self.H_VoigtSize = H_Voigt.shape[0]


        return H_Voigt



    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        alpha = self.alpha
        beta = self.beta
        kappa = self.kappa

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        F = StrainTensors['F'][gcounter]
        H = J*np.linalg.inv(F).T
        g = np.dot(H,H.T)
        bcross = trace(b)*b-np.dot(b,b)
        # b=np.dot(F,F.T)



        # stress = 2.*alpha*J**(-5/3.)*b - 2./3.*alpha*J**(-5/3.)*trace(b)*I + \
        #       beta*J**(-3)*trace(g)**(3./2.)*I - 3*beta*J**(-3)*trace(g)**(1./2.)*g + \
        #       +(kappa*(J-1.0))*I #####

        if self.ndim == 2:
            trb = trace(b)+1
            trg = trace(g)+J**2
        elif self.ndim == 3:
            trb = trace(b)
            trg = trace(g)

        stress = 2.*alpha*J**(-5/3.)*b - 2./3.*alpha*J**(-5/3.)*(trb)*I + \
                beta*J**(-3)*(trg)**(3./2.)*I - 3*beta*J**(-3)*(trg)**(1./2.)*g + \
                +(kappa*(J-1.0))*I 

        return stress



from copy import deepcopy
import gc
from numpy.linalg import det, inv, norm, cond
from Florence import QuadratureRule, FunctionSpace
from Florence.FiniteElements.LocalAssembly.KinematicMeasures import *
from Florence.FiniteElements.LocalAssembly._KinematicMeasures_ import _KinematicMeasures_
from ._ConstitutiveStiffnessDF_ import __ConstitutiveStiffnessIntegrandDF__
from Florence.Tensor import issymetric
from Florence.LegendreTransform import LegendreTransform
from scipy.sparse import coo_matrix, csc_matrix, csr_matrix
from .VariationalPrinciple import *

from Florence.FiniteElements.Assembly.SparseAssemblyNative import SparseAssemblyNative
from Florence.FiniteElements.Assembly.RHSAssemblyNative import RHSAssemblyNative



__all__ = ["CoupleStressFormulation"]

class CoupleStressFormulation(VariationalPrinciple):

    def __init__(self, mesh, variables_order=(1,0,0), subtype="lagrange_multiplier",
        quadrature_rules=None, quadrature_type=None, function_spaces=None, compute_post_quadrature=False,
        equally_spaced_bases=False, save_condensed_matrices=True):
        """

            Input:
                subtype:                    [str] either "lagrange_multiplier", "augmented_lagrange" or "penalty"
        """

        if mesh.element_type != "tet" and mesh.element_type != "tri" and \
            mesh.element_type != "quad" and mesh.element_type != "hex":
            raise NotImplementedError( type(self).__name__, "has not been implemented for", mesh.element_type, "elements")

        if isinstance(variables_order,int):
            self.variables_order = (self.variables_order,)
        self.variables_order = variables_order

        super(CoupleStressFormulation, self).__init__(mesh,variables_order=self.variables_order,
            quadrature_type=quadrature_type,quadrature_rules=quadrature_rules,function_spaces=function_spaces,
            compute_post_quadrature=compute_post_quadrature)

        self.fields = "couple_stress"
        self.nvar = self.ndim
        self.subtype = subtype
        self.save_condensed_matrices = save_condensed_matrices

        C = mesh.InferPolynomialDegree() - 1
        mesh.InferBoundaryElementType()

        if C < 1:
            raise ValueError("Incorrect initial mesh provided for the formulation. Mesh has to be at least order 2")

        # CHECK IF MESH IS APPROPRIATE
        # if C == 0:
            # warn('Mesh not appropriate for formulation')
            # raise ValueError('Mesh not appropriate for formulation. p>1 for primary variable (displacement)')
        # BUILD MESHES FOR ALL FIELDS
        p = C+1
        # DISPLACEMENTS
        mesh0 = deepcopy(mesh)
        # ROTATIONS
        mesh1 = deepcopy(mesh)
        mesh1 = mesh1.GetLinearMesh(remap=True)
        mesh1.GetHighOrderMesh(p=p-1)
        # LAGRANGE MULTIPLIER
        mesh2 = deepcopy(mesh)
        mesh2 = mesh2.GetLinearMesh(remap=True)
        mesh2.GetHighOrderMesh(p=p-1)
        # ALL MESHES
        self.meshes = (mesh0,mesh1,mesh2)


        # GET QUADRATURE RULES
        norder = C+2
        if mesh.element_type == "quad" or mesh.element_type == "hex":
            norder = C+1

        if quadrature_rules == None and self.quadrature_rules == None:
            # FOR DISPLACEMENT
            quadrature0 = QuadratureRule(optimal=3, norder=self.GetQuadratureOrder(norder,mesh.element_type)[0],
                mesh_type=mesh.element_type, is_flattened=False)
            # FOR ROTATIONS
            quadrature1 = QuadratureRule(optimal=3, norder=self.GetQuadratureOrder(norder,mesh.element_type)[0],
                mesh_type=mesh.element_type, is_flattened=False)
            # FOR LAGRANGE MULTIPLIER
            quadrature2 = QuadratureRule(optimal=3, norder=self.GetQuadratureOrder(norder,mesh.element_type)[0],
                mesh_type=mesh.element_type, is_flattened=False)
            # BOUNDARY
            bquadrature = QuadratureRule(optimal=3, norder=C+2, mesh_type=mesh.boundary_element_type, is_flattened=False)

            self.quadrature_rules = (quadrature0,quadrature1,quadrature2,bquadrature)
        else:
            self.quadrature_rules = quadrature_rules


        # GET FUNCTIONAL SPACES
        if function_spaces == None and self.function_spaces == None:
            # FOR DISPLACEMENT
            function_space0 = FunctionSpace(mesh0, self.quadrature_rules[0], p=mesh0.degree,
                equally_spaced=equally_spaced_bases, use_optimal_quadrature=False)
            # FOR ROTATIONS
            function_space1 = FunctionSpace(mesh1, self.quadrature_rules[1], p=mesh1.degree,
                equally_spaced=equally_spaced_bases, use_optimal_quadrature=False)
            # FOR LAGRANGE MULTIPLIER
            function_space2 = FunctionSpace(mesh2, self.quadrature_rules[2], p=mesh2.degree,
                equally_spaced=equally_spaced_bases, use_optimal_quadrature=False)
            # BOUNDARY
            bfunction_space = FunctionSpace(mesh0.CreateDummyLowerDimensionalMesh(), self.quadrature_rules[3], p=mesh0.degree,
                equally_spaced=equally_spaced_bases, use_optimal_quadrature=False)

            self.function_spaces = (function_space0, function_space1, function_space2, bfunction_space)
        else:
            self.function_spaces = function_spaces


        # local_size = function_space.Bases.shape[0]*self.nvar
        local_size = self.function_spaces[0].Bases.shape[0]*self.nvar
        self.local_rows = np.repeat(np.arange(0,local_size),local_size,axis=0)
        self.local_columns = np.tile(np.arange(0,local_size),local_size)
        self.local_size = local_size

        # FOR MASS
        local_size_m = self.function_spaces[0].Bases.shape[0]*self.nvar
        self.local_rows_mass = np.repeat(np.arange(0,local_size_m),local_size_m,axis=0)
        self.local_columns_mass = np.tile(np.arange(0,local_size_m),local_size_m)
        self.local_size_m = local_size_m


        if self.save_condensed_matrices:
            # elist = [0]*mesh.nelem # CANT USE ONE PRE-CREATED LIST AS IT GETS MODIFIED
            # KEEP VECTORS AND MATRICES SEPARATE BECAUSE OF THE SAME REASON
            if self.subtype == "lagrange_multiplier":
                self.condensed_matrices = {'k_uu':[0]*mesh.nelem,'k_us':[0]*mesh.nelem,
                'k_ww':[0]*mesh.nelem,'k_ws':[0]*mesh.nelem,'inv_k_ws':[0]*mesh.nelem}
                self.condensed_vectors = {'tu':[0]*mesh.nelem,'tw':[0]*mesh.nelem,'ts':[0]*mesh.nelem}
            elif self.subtype == "augmented_lagrange":
                self.condensed_matrices = {'k_uu':[0]*mesh.nelem,'k_us':[0]*mesh.nelem,
                'k_ww':[0]*mesh.nelem,'k_ws':[0]*mesh.nelem,'k_ss':[0]*mesh.nelem,'inv_k_ws':[0]*mesh.nelem}
                self.condensed_vectors = {'tu':[0]*mesh.nelem,'tw':[0]*mesh.nelem,'ts':[0]*mesh.nelem}
            elif self.subtype == "penalty":
                self.condensed_matrices = {'k_uu':[0]*mesh.nelem,'k_uw':[0]*mesh.nelem,'k_ww':[0]*mesh.nelem}
                self.condensed_vectors = {'tu':[0]*mesh.nelem,'tw':[0]*mesh.nelem}

        # COMPUTE THE COMMON/NEIGHBOUR NODES ONCE
        self.all_nodes = np.unique(self.meshes[1].elements)
        self.Elss, self.Poss = self.meshes[1].GetNodeCommonality()[:2]


    def GetElementalMatrices(self, elem, function_space, mesh, material, fem_solver, Eulerx, Eulerw, Eulers, Eulerp):

        massel=[]; f = []
        # COMPUTE THE STIFFNESS MATRIX
        if material.has_low_level_dispatcher:
            stiffnessel, t = self.__GetLocalStiffness__(material, fem_solver, Eulerx, Eulerw, Eulers, Eulerp, elem)
        else:
            stiffnessel, t = self.GetLocalStiffness(material, fem_solver, Eulerx, Eulerw, Eulers, Eulerp, elem)

        I_mass_elem = []; J_mass_elem = []; V_mass_elem = []
        if fem_solver.analysis_type != 'static' and fem_solver.is_mass_computed is False:
            # GET THE FIELDS AT THE ELEMENT LEVEL
            LagrangeElemCoords = mesh[0].points[mesh[0].elements[elem,:],:]
            EulerElemCoords = Eulerx[mesh[0].elements[elem,:],:]
            # COMPUTE THE MASS MATRIX
            if material.has_low_level_dispatcher:
                massel = self.__GetLocalMass__(material,fem_solver,elem)
            else:
                # massel = self.GetLocalMass(material,fem_solver,elem)
                massel = self.GetLocalMass(function_space[0], material, LagrangeElemCoords, EulerElemCoords, fem_solver, elem)


        if fem_solver.has_moving_boundary:
            # COMPUTE FORCE VECTOR
            f = ApplyNeumannBoundaryConditions3D(MainData, mesh, elem, LagrangeElemCoords)

        I_stiff_elem, J_stiff_elem, V_stiff_elem = self.FindIndices(stiffnessel)
        if fem_solver.analysis_type != 'static' and fem_solver.is_mass_computed is False:
            I_mass_elem, J_mass_elem, V_mass_elem = self.FindIndices(massel)

        return I_stiff_elem, J_stiff_elem, V_stiff_elem, t, f, I_mass_elem, J_mass_elem, V_mass_elem


    def GetMassMatrix(self, elem, function_space, mesh, material, fem_solver, Eulerx, Eulerw, Eulerp):

        massel=[]
        # COMPUTE THE MASS MATRIX
        # if material.has_low_level_dispatcher:
        #     massel = self.__GetLocalMass__(material,fem_solver,elem)
        # else:
        #     massel = self.GetLocalMass(material,fem_solver,elem)
        massel = self.__GetLocalMass__(material,fem_solver,elem)

        I_mass_elem, J_mass_elem, V_mass_elem = self.FindIndices(massel)

        return I_mass_elem, J_mass_elem, V_mass_elem


    def GetLocalStiffness(self, material, fem_solver, Eulerx, Eulerw, Eulers, Eulerp=None, elem=0):
        """Get stiffness matrix of the system"""

        # return self.K_uu(material, fem_solver, Eulerx, Eulerp, elem=0)
        if self.subtype=="lagrange_multiplier" or self.subtype=="augmented_lagrange":

            tractionforce = []
            k_uu, tu = self.K_uu(material, fem_solver, Eulerx, Eulerp, elem)
            k_uw = self.K_uw(material, fem_solver, Eulerx, Eulerp, elem)
            k_us = self.K_us(material, fem_solver, Eulerx, Eulerp, elem)


            # k_ww, tw = self.K_ww(material, fem_solver, Eulerw, Eulerp, elem)
            k_ww, tw = self.K_ww(material, fem_solver, Eulerx, Eulerp, elem)    # CHECK Eulerx vs Eulerw
            k_ws = self.K_ws(material, fem_solver, Eulerw, Eulerp, elem)

            k_ss, ts = self.K_ss(material, fem_solver, Eulerw, Eulerp, elem)

            if fem_solver.static_condensation is True:
                # IF STATIC CONDENSATION
                if self.subtype=="lagrange_multiplier":
                    inv_k_ws = inv(k_ws)
                    k0 = k_ww.dot(inv_k_ws)
                    k1 = k0.dot(k_us.T)
                    stiffness = k_uu + np.dot(np.dot(k_us,inv_k_ws),k1)

                    t0 = tw - np.dot(k0,ts)
                    tractionforce = tu - np.dot(np.dot(k_us,inv_k_ws),t0)

                    if self.save_condensed_matrices:
                        self.condensed_matrices['k_uu'][elem] = k_uu
                        self.condensed_matrices['k_us'][elem] = k_us
                        self.condensed_matrices['k_ww'][elem] = k_ww
                        self.condensed_matrices['k_ws'][elem] = k_ws
                        self.condensed_matrices['inv_k_ws'][elem] = inv_k_ws
                        self.condensed_vectors['tu'][elem] = tu
                        self.condensed_vectors['tw'][elem] = tw
                        self.condensed_vectors['ts'][elem] = ts

                elif self.subtype=="augmented_lagrange":
                    inv_k_ws = inv(k_ws)
                    k0 = k_ww.dot(inv_k_ws)
                    k1 = k0.dot(k_us.T)
                    k2 = inv(k_ws - k0.dot(k_ss))
                    stiffness = k_uu + np.dot(np.dot(k_us,k2),k1)

                    t0 = tw - np.dot(k0,ts)
                    tractionforce = tu - np.dot(np.dot(k_us,k2),t0)

                    if self.save_condensed_matrices:
                        self.condensed_matrices['k_uu'][elem] = k_uu
                        self.condensed_matrices['k_us'][elem] = k_us
                        self.condensed_matrices['k_ww'][elem] = k_ww
                        self.condensed_matrices['k_ws'][elem] = k_ws
                        self.condensed_matrices['k_ss'][elem] = k_ss
                        self.condensed_matrices['inv_k_ws'][elem] = inv_k_ws
                        self.condensed_vectors['tu'][elem] = tu
                        self.condensed_vectors['tw'][elem] = tw
                        self.condensed_vectors['ts'][elem] = ts

            else:
                # IF NO STATIC CONDENSATION
                raise NotImplementedError("Not implemented yet")
                k0 = np.concatenate((k_uu,k_uw, k_us),axis=1)
                k1 = np.concatenate((k_uw.T,k_ww, k_ws),axis=1)
                k2 = np.concatenate((k_us.T,k_ws.T, k_ss),axis=1)
                stiffness = np.concatenate((k0,k1, k2),axis=0)
                tractionforce = np.concatenate((tu,tw,ts))



        elif self.subtype=="penalty":

            tractionforce = []
            k_uu, tu = self.K_uu(material, fem_solver, Eulerx, Eulerp, elem)
            k_uu2, tu2 = self.K_uu_Penalty(material, fem_solver, Eulerx, Eulerp, elem)
            k_uw = material.kappa*self.K_us(material, fem_solver, Eulerx, Eulerp, elem)
            k_ww, tw = self.K_ww_Penalty(material, fem_solver, Eulerw, Eulerp, elem)

            if fem_solver.static_condensation is True:
                # IF STATIC CONDENSATION
                inv_k_ww = inv(k_ww)
                # stiffness = k_uu - np.dot(np.dot(k_uw,inv_k_ww),k_uw.T)
                stiffness = k_uu + k_uu2 - np.dot(np.dot(k_uw,inv_k_ww),k_uw.T)
                tractionforce = tu + tu2 - np.dot(np.dot(k_uw,inv_k_ww),tw)
            else:
                # IF NO STATIC CONDENSATION
                raise NotImplementedError("Not implemented yet")
                k0 = np.concatenate((k_uu+k_uu2,-k_uw),axis=1)
                stiffness = np.concatenate((-k_uw.T,-k_ww),axis=1)
                tractionforce = np.concatenate((tu,tw))


        else:
            raise ValueError("subtype of this variational formulation should be 'lagrange_multiplier' or 'penalty'")



        return stiffness, tractionforce



    def K_uu(self, material, fem_solver, Eulerx, Eulerp=None, elem=0):
        """Get stiffness matrix of the system"""

        meshes = self.meshes
        mesh = self.meshes[0]
        function_spaces = self.function_spaces
        function_space = self.function_spaces[0]

        ndim = self.ndim
        nvar = self.nvar
        nodeperelem = meshes[0].elements.shape[1]
        # print nodeperelem

        # GET THE FIELDS AT THE ELEMENT LEVEL
        LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
        EulerELemCoords = Eulerx[mesh.elements[elem,:],:]

        Jm = function_spaces[0].Jm
        AllGauss = function_space.AllGauss


        # GET LOCAL KINEMATICS
        SpatialGradient, F, detJ = _KinematicMeasures_(Jm, AllGauss[:,0],
            LagrangeElemCoords, EulerELemCoords, fem_solver.requires_geometry_update)
        # COMPUTE WORK-CONJUGATES AND HESSIAN AT THIS GAUSS POINT
        CauchyStressTensor, H_Voigt, _ = material.KineticMeasures(F,elem=elem)
        # COMPUTE LOCAL CONSTITUTIVE STIFFNESS AND TRACTION
        stiffness, tractionforce = __ConstitutiveStiffnessIntegrandDF__(SpatialGradient,
            CauchyStressTensor,H_Voigt,detJ,self.nvar,fem_solver.requires_geometry_update)
        # # COMPUTE GEOMETRIC STIFFNESS
        # if fem_solver.requires_geometry_update:
        #     stiffness += self.__GeometricStiffnessIntegrand__(SpatialGradient,CauchyStressTensor,detJ)

        # SAVE AT THIS GAUSS POINT
        self.SpatialGradient = SpatialGradient
        self.detJ = detJ

        return stiffness, tractionforce

        # # ALLOCATE
        # stiffness = np.zeros((nodeperelem*nvar,nodeperelem*nvar),dtype=np.float64)
        # tractionforce = np.zeros((nodeperelem*nvar,1),dtype=np.float64)
        # B = np.zeros((nodeperelem*nvar,material.elasticity_tensor_size),dtype=np.float64)

        # # COMPUTE KINEMATIC MEASURES AT ALL INTEGRATION POINTS USING EINSUM (AVOIDING THE FOR LOOP)
        # # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
        # ParentGradientX = np.einsum('ijk,jl->kil', Jm, LagrangeElemCoords)
        # # MATERIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla_0 (N)]
        # MaterialGradient = np.einsum('ijk,kli->ijl', inv(ParentGradientX), Jm)
        # # DEFORMATION GRADIENT TENSOR [\vec{x} \otimes \nabla_0 (N)]
        # F = np.einsum('ij,kli->kjl', EulerELemCoords, MaterialGradient)

        # # COMPUTE REMAINING KINEMATIC MEASURES
        # StrainTensors = KinematicMeasures(F, fem_solver.analysis_nature)

        # # UPDATE/NO-UPDATE GEOMETRY
        # if fem_solver.requires_geometry_update:
        #     # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
        #     ParentGradientx = np.einsum('ijk,jl->kil',Jm,EulerELemCoords)
        #     # SPATIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla (N)]
        #     SpatialGradient = np.einsum('ijk,kli->ilj',inv(ParentGradientx),Jm)
        #     # COMPUTE ONCE detJ (GOOD SPEEDUP COMPARED TO COMPUTING TWICE)
        #     detJ = np.einsum('i,i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)),np.abs(StrainTensors['J']))
        # else:
        #     # SPATIAL GRADIENT AND MATERIAL GRADIENT TENSORS ARE EQUAL
        #     SpatialGradient = np.einsum('ikj',MaterialGradient)
        #     # COMPUTE ONCE detJ
        #     detJ = np.einsum('i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)))


        # # LOOP OVER GAUSS POINTS
        # for counter in range(AllGauss.shape[0]):

        #     # COMPUTE THE HESSIAN AT THIS GAUSS POINT
        #     material.Hessian(StrainTensors,None, elem, counter)
        #     H_Voigt = material.elasticity_tensor

        #     # COMPUTE CAUCHY STRESS TENSOR
        #     CauchyStressTensor = []
        #     if fem_solver.requires_geometry_update:
        #         CauchyStressTensor = material.CauchyStress(StrainTensors,None,elem,counter)


        #     # COMPUTE THE TANGENT STIFFNESS MATRIX
        #     BDB_1, t = self.K_uu_Integrand(B, SpatialGradient[counter,:,:],
        #         None, CauchyStressTensor, H_Voigt, analysis_nature=fem_solver.analysis_nature,
        #         has_prestress=fem_solver.has_prestress)

        #     # COMPUTE GEOMETRIC STIFFNESS MATRIX
        #     if fem_solver.requires_geometry_update:
        #         # BDB_1 += self.GeometricStiffnessIntegrand(SpatialGradient[counter,:,:],CauchyStressTensor)
        #         # INTEGRATE TRACTION FORCE
        #         tractionforce += t*detJ[counter]

        #     # INTEGRATE STIFFNESS
        #     stiffness += BDB_1*detJ[counter]


        # # SAVE AT THIS GAUSS POINT
        # self.SpatialGradient = SpatialGradient
        # self.detJ = detJ

        # return stiffness, tractionforce


    def K_uw(self, material, fem_solver, Eulerx, Eulerp=None, elem=0):
        """Get stiffness matrix of the system"""
        return np.zeros((self.meshes[0].elements.shape[1]*self.ndim,self.meshes[1].elements.shape[1]*self.ndim),dtype=np.float64)


    def K_us(self, material, fem_solver, Eulerx, Eulerp=None, elem=0):
        """Get stiffness matrix of the system"""

        meshes = self.meshes
        function_spaces = self.function_spaces

        Bases_s = function_spaces[2].Bases
        Ns = np.zeros((self.ndim,Bases_s.shape[0]*self.ndim),dtype=np.float64)
        Bu = np.zeros((self.meshes[0].elements.shape[1]*self.ndim,self.ndim),dtype=np.float64)
        stiffness = np.zeros((self.meshes[0].elements.shape[1]*self.ndim,self.meshes[2].elements.shape[1]*self.ndim))

        AllGauss = function_spaces[0].AllGauss
        # LOOP OVER GAUSS POINTS
        for counter in range(AllGauss.shape[0]):

            # COMPUTE THE TANGENT STIFFNESS MATRIX
            Bu_Ns = self.K_us_Integrand(Bu, Ns, self.SpatialGradient[counter,:,:], Bases_s[:,counter])
            # INTEGRATE STIFFNESS
            stiffness += Bu_Ns*self.detJ[counter]


        return stiffness


    def K_ww(self, material, fem_solver, Eulerw, Eulerp=None, elem=0):
        """Get stiffness matrix of the system"""

        meshes = self.meshes
        mesh = self.meshes[1]

        function_spaces = self.function_spaces
        function_space = self.function_spaces[1]

        ndim = self.ndim
        nvar = ndim
        nodeperelem = meshes[1].elements.shape[1]

        # GET THE FIELDS AT THE ELEMENT LEVEL
        LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
        EulerELemCoords = Eulerw[mesh.elements[elem,:],:]

        Jm = function_spaces[1].Jm
        AllGauss = function_space.AllGauss

        # # GET LOCAL KINEMATICS
        # SpatialGradient, F, detJ = _KinematicMeasures_(Jm, AllGauss[:,0],
        #     LagrangeElemCoords, EulerELemCoords, fem_solver.requires_geometry_update)
        # # COMPUTE WORK-CONJUGATES AND HESSIAN AT THIS GAUSS POINT
        # CauchyStressTensor, _, H_Voigt = material.KineticMeasures(F,elem=elem)


        # ALLOCATE
        stiffness = np.zeros((nodeperelem*nvar,nodeperelem*nvar),dtype=np.float64)
        tractionforce = np.zeros((nodeperelem*nvar,1),dtype=np.float64)
        B = np.zeros((nodeperelem*nvar,material.gradient_elasticity_tensor_size),dtype=np.float64)

        # COMPUTE KINEMATIC MEASURES AT ALL INTEGRATION POINTS USING EINSUM (AVOIDING THE FOR LOOP)
        # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
        ParentGradientX = np.einsum('ijk,jl->kil', Jm, LagrangeElemCoords)
        # MATERIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla_0 (N)]
        MaterialGradient = np.einsum('ijk,kli->ijl', inv(ParentGradientX), Jm)
        # DEFORMATION GRADIENT TENSOR [\vec{x} \otimes \nabla_0 (N)]
        F = np.einsum('ij,kli->kjl', EulerELemCoords, MaterialGradient)

        # COMPUTE REMAINING KINEMATIC MEASURES
        StrainTensors = KinematicMeasures(F, fem_solver.analysis_nature)

        # UPDATE/NO-UPDATE GEOMETRY
        if fem_solver.requires_geometry_update:
            # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
            ParentGradientx = np.einsum('ijk,jl->kil',Jm,EulerELemCoords)
            # SPATIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla (N)]
            SpatialGradient = np.einsum('ijk,kli->ilj',inv(ParentGradientx),Jm)
            # COMPUTE ONCE detJ (GOOD SPEEDUP COMPARED TO COMPUTING TWICE)
            detJ = np.einsum('i,i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)),np.abs(StrainTensors['J']))
        else:
            # SPATIAL GRADIENT AND MATERIAL GRADIENT TENSORS ARE EQUAL
            SpatialGradient = np.einsum('ikj',MaterialGradient)
            # COMPUTE ONCE detJ
            detJ = np.einsum('i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)))


        # LOOP OVER GAUSS POINTS
        for counter in range(AllGauss.shape[0]):

            # COMPUTE THE HESSIAN AT THIS GAUSS POINT
            material.Hessian(StrainTensors,None, elem, counter)
            H_Voigt = material.gradient_elasticity_tensor

            # COMPUTE CAUCHY STRESS TENSOR
            CoupleStressVector = []
            if fem_solver.requires_geometry_update:
                CoupleStressVector = material.CoupleStress(StrainTensors,None,elem,counter).reshape(self.ndim,1)


            # COMPUTE THE TANGENT STIFFNESS MATRIX
            BDB_1, t = self.K_ww_Integrand(B, SpatialGradient[counter,:,:],
                None, CoupleStressVector, H_Voigt, analysis_nature=fem_solver.analysis_nature,
                has_prestress=fem_solver.has_prestress)

            # COMPUTE GEOMETRIC STIFFNESS MATRIX
            if fem_solver.requires_geometry_update:
                # INTEGRATE TRACTION FORCE
                tractionforce += t*detJ[counter]

            # INTEGRATE STIFFNESS
            stiffness += BDB_1*detJ[counter]


        # # SAVE AT THIS GAUSS POINT
        # self.SpatialGradient = SpatialGradient
        # self.detJ = detJ

        return stiffness, tractionforce


    def K_ws(self, material, fem_solver, Eulerw, Eulerp=None, elem=0):
        """Get stiffness matrix of the system"""

        meshes = self.meshes
        function_spaces = self.function_spaces

        Bases_w = function_spaces[1].Bases
        Bases_s = function_spaces[2].Bases
        Nw = np.zeros((Bases_w.shape[0]*self.ndim,self.ndim),dtype=np.float64)
        Ns = np.zeros((self.ndim,Bases_s.shape[0]*self.ndim),dtype=np.float64)
        stiffness = np.zeros((Bases_w.shape[0]*self.ndim,Bases_s.shape[0]*self.ndim))

        AllGauss = function_spaces[0].AllGauss
        # LOOP OVER GAUSS POINTS
        for counter in range(AllGauss.shape[0]):

            # COMPUTE THE TANGENT STIFFNESS MATRIX
            Nw_Ns = self.K_ws_Integrand(Nw, Ns, Bases_w[:,counter], Bases_s[:,counter])
            # INTEGRATE STIFFNESS
            stiffness += Nw_Ns*self.detJ[counter]  ## CAREFUL ABOUT [CHECK] self.detJ[counter] ####################


        return -stiffness


    def K_ss(self, material, fem_solver, Eulers, Eulerp=None, elem=0):
        """Get stiffness matrix of the system"""
        stiffness = np.zeros((self.function_spaces[2].Bases.shape[0]*self.ndim,self.function_spaces[2].Bases.shape[0]*self.ndim),dtype=np.float64)
        tractionforce = np.zeros((self.function_spaces[2].Bases.shape[0]*self.ndim,1),dtype=np.float64)
        if self.subtype == "lagrange_multiplier":
            return stiffness, tractionforce

        EulerELemS = Eulers[self.meshes[2].elements[elem,:],:]
        Bases_s = self.function_spaces[2].Bases
        Ns = np.zeros((self.ndim,Bases_s.shape[0]*self.ndim),dtype=np.float64)
        AllGauss = self.function_spaces[2].AllGauss

        # FIND LAGRANGE MULTIPLIER AT ALL GAUSS POINTS
        EulerGaussS = np.dot(Bases_s.T,EulerELemS)

        # LOOP OVER GAUSS POINTS
        for counter in range(AllGauss.shape[0]):

            # COMPUTE STRESS
            LagrangeMultiplierStressVector = material.LagrangeMultiplierStress(EulerGaussS,elem=elem,gcounter=counter)
            # COMPUTE THE TANGENT STIFFNESS MATRIX
            NDN, t = self.K_ss_Integrand(Ns, Bases_s[:,counter], 0, LagrangeMultiplierStressVector, material.kappa,
                analysis_nature=fem_solver.analysis_nature, has_prestress=fem_solver.has_prestress)
            # INTEGRATE STIFFNESS
            stiffness += NDN*self.detJ[counter]  ## CAREFUL ABOUT [CHECK] self.detJ[counter] ####################
            # INTEGRAGE TRACTION
            if fem_solver.requires_geometry_update:
                # INTEGRATE TRACTION FORCE
                tractionforce += t*detJ[counter]

        return stiffness, tractionforce



    def K_uu_Penalty(self, material, fem_solver, Eulerx, Eulerp=None, elem=0):
        """Get stiffness matrix of the system"""

        meshes = self.meshes
        function_spaces = self.function_spaces

        Bu = np.zeros((self.meshes[0].elements.shape[1]*self.ndim,self.ndim),dtype=np.float64)
        stiffness = np.zeros((self.meshes[0].elements.shape[1]*self.ndim,self.meshes[0].elements.shape[1]*self.ndim))

        AllGauss = function_spaces[0].AllGauss
        # LOOP OVER GAUSS POINTS
        for counter in range(AllGauss.shape[0]):

            # COMPUTE THE TANGENT STIFFNESS MATRIX
            BDB = self.K_uu_Penalty_Integrand(Bu, self.SpatialGradient[counter,:,:])
            # INTEGRATE STIFFNESS
            stiffness += material.kappa*BDB*self.detJ[counter]


        # THIS CONTRIBUTES TO TRACTION AS WELL
        tractionforce = np.zeros((self.meshes[0].elements.shape[1]*self.ndim,1))
        return stiffness, tractionforce


    def K_ww_Penalty(self, material, fem_solver, Eulerw, Eulerp=None, elem=0):
        """Get stiffness matrix of the system"""

        meshes = self.meshes
        mesh = self.meshes[1]

        function_spaces = self.function_spaces
        function_space = self.function_spaces[1]

        ndim = self.ndim
        nvar = self.ndim
        nodeperelem = meshes[1].elements.shape[1]

        Jm = function_spaces[1].Jm
        AllGauss = function_space.AllGauss

        # ALLOCATE
        stiffness = np.zeros((nodeperelem*nvar,nodeperelem*nvar),dtype=np.float64)
        tractionforce = np.zeros((nodeperelem*nvar,1),dtype=np.float64)
        Bases_w = self.function_spaces[1].Bases
        Nw = np.zeros((self.ndim,Bases_w.shape[0]*self.ndim),dtype=np.float64)
        # detJ = AllGauss[:,0]
        detJ = self.detJ

        # LOOP OVER GAUSS POINTS
        for counter in range(AllGauss.shape[0]):

            # COMPUTE CAUCHY STRESS TENSOR
            CoupleStressVector = []
            if fem_solver.requires_geometry_update:
                CoupleStressVector = material.CoupleStress(StrainTensors,None,elem,counter).reshape(self.ndim,1)


            # COMPUTE THE TANGENT STIFFNESS MATRIX
            BDB_1, t = self.K_ww_Penalty_Integrand(Nw, Bases_w[:,counter],
                0, CoupleStressVector, material.kappa, analysis_nature=fem_solver.analysis_nature,
                has_prestress=fem_solver.has_prestress)

            # COMPUTE GEOMETRIC STIFFNESS MATRIX
            if fem_solver.requires_geometry_update:
                # INTEGRATE TRACTION FORCE
                tractionforce += t*detJ[counter]

            # INTEGRATE STIFFNESS
            stiffness += material.kappa*BDB_1*detJ[counter]


        return stiffness, tractionforce


    def GetLocalTraction(self, function_space, material, LagrangeElemCoords,
        EulerELemCoords, ElectricPotentialElem, fem_solver, elem=0):
        """Get traction vector of the system"""
        pass

    def K_uu_Integrand(self, B, SpatialGradient, ElectricDisplacementx,
        CauchyStressTensor, H_Voigt, analysis_nature="nonlinear", has_prestress=True):

        ndim = self.ndim
        nvar = self.nvar

        # MATRIX FORM
        SpatialGradient = SpatialGradient.T

        # THREE DIMENSIONS
        if SpatialGradient.shape[0]==3:

            B[0::nvar,0] = SpatialGradient[0,:]
            B[1::nvar,1] = SpatialGradient[1,:]
            B[2::nvar,2] = SpatialGradient[2,:]
            # Mechanical - Shear Terms
            B[1::nvar,5] = SpatialGradient[2,:]
            B[2::nvar,5] = SpatialGradient[1,:]

            B[0::nvar,4] = SpatialGradient[2,:]
            B[2::nvar,4] = SpatialGradient[0,:]

            B[0::nvar,3] = SpatialGradient[1,:]
            B[1::nvar,3] = SpatialGradient[0,:]

            if analysis_nature == 'nonlinear' or has_prestress:
                CauchyStressTensor_Voigt = np.array([
                    CauchyStressTensor[0,0],CauchyStressTensor[1,1],CauchyStressTensor[2,2],
                    CauchyStressTensor[0,1],CauchyStressTensor[0,2],CauchyStressTensor[1,2]
                    ]).reshape(6,1)

                TotalTraction = CauchyStressTensor_Voigt

        elif SpatialGradient.shape[0]==2:

            B[0::nvar,0] = SpatialGradient[0,:]
            B[1::nvar,1] = SpatialGradient[1,:]
            # Mechanical - Shear Terms
            B[0::nvar,2] = SpatialGradient[1,:]
            B[1::nvar,2] = SpatialGradient[0,:]

            if analysis_nature == 'nonlinear' or has_prestress:
                CauchyStressTensor_Voigt = np.array([
                    CauchyStressTensor[0,0],CauchyStressTensor[1,1],
                    CauchyStressTensor[0,1]]).reshape(3,1)

                TotalTraction = CauchyStressTensor

        BDB = np.dot(np.dot(B,H_Voigt),B.T)
        t=[]
        if analysis_nature == 'nonlinear' or has_prestress:
            t = np.dot(B,TotalTraction)

        return BDB, t


    def K_us_Integrand(self, Bu, Ns, SpatialGradient, Bases_s):

        ndim = self.ndim
        nvar = self.nvar

        # MATRIX FORM
        SpatialGradient = SpatialGradient.T

        # THREE DIMENSIONS
        if SpatialGradient.shape[0]==3:

            # VORTICITY TERMS
            Bu[1::nvar,0] = -SpatialGradient[2,:]
            Bu[2::nvar,0] = SpatialGradient[1,:]

            Bu[0::nvar,1] = SpatialGradient[2,:]
            Bu[2::nvar,1] = -SpatialGradient[0,:]

            Bu[0::nvar,2] = -SpatialGradient[1,:]
            Bu[1::nvar,2] = SpatialGradient[0,:]

        elif SpatialGradient.shape[0]==2:
            # VORTICITY TERMS
            Bu[0::nvar,0] = -SpatialGradient[1,:]
            Bu[1::nvar,0] = SpatialGradient[0,:]

        for ivar in range(ndim):
            Ns[ivar,ivar::nvar] = Bases_s

        Bu_Ns = 0.5*np.dot(Bu,Ns)
        return Bu_Ns


    def K_ww_Integrand(self, B, SpatialGradient, ElectricDisplacementx,
        CoupleStressVector, H_Voigt, analysis_nature="nonlinear", has_prestress=True):

        ndim = self.ndim
        nvar = self.nvar

        # MATRIX FORM
        SpatialGradient = SpatialGradient.T

        # THREE DIMENSIONS
        if SpatialGradient.shape[0]==3:

            # VORTICITY TERMS
            B[1::nvar,0] = -SpatialGradient[2,:]
            B[2::nvar,0] = SpatialGradient[1,:]

            B[0::nvar,1] = SpatialGradient[2,:]
            B[2::nvar,1] = -SpatialGradient[0,:]

            B[0::nvar,2] = -SpatialGradient[1,:]
            B[1::nvar,2] = SpatialGradient[0,:]

        elif SpatialGradient.shape[0]==2:
            # VORTICITY TERMS
            B[0::nvar,0] = -SpatialGradient[1,:]
            B[1::nvar,0] = SpatialGradient[0,:]


        BDB = np.dot(np.dot(B,H_Voigt),B.T)
        t=[]
        if analysis_nature == 'nonlinear' or has_prestress:
            t = np.dot(B,CoupleStressVector)
        return BDB, t


    def K_ws_Integrand(self, Nw, Ns, Bases_w, Bases_s):

        ndim = self.ndim
        nvar = self.nvar

        for ivar in range(ndim):
            Nw[ivar::nvar,ivar] = Bases_w

        for ivar in range(ndim):
            Ns[ivar,ivar::nvar] = Bases_s

        Nw_Ns = 0.5*np.dot(Nw,Ns)
        return Nw_Ns


    def K_ss_Integrand(self, Ns, Bases_s, ElectricDisplacementx,
        LagrangeMultiplierStressVector, kappa, analysis_nature="nonlinear", has_prestress=True):

        ndim = self.ndim
        nvar = self.nvar

        for ivar in range(ndim):
            Ns[ivar,ivar::nvar] = Bases_s

        if self.subtype == "augmented_lagrange":
            NDN = np.dot(Ns.T,Ns)/(1.0*kappa)
        else:
            NDN = np.zeros((self.function_spaces[2].Bases.shape[0]*self.ndim,self.function_spaces[2].Bases.shape[0]*self.ndim),dtype=np.float64)
        t=[]
        if analysis_nature == 'nonlinear' or has_prestress:
            t = np.dot(Ns,LagrangeMultiplierStressVector)
        return NDN, t


    def K_uu_Penalty_Integrand(self, Bu, SpatialGradient):

        ndim = self.ndim
        nvar = self.nvar

        # MATRIX FORM
        SpatialGradient = SpatialGradient.T

        # THREE DIMENSIONS
        if SpatialGradient.shape[0]==3:

            # VORTICITY TERMS
            Bu[1::nvar,0] = -SpatialGradient[2,:]
            Bu[2::nvar,0] = SpatialGradient[1,:]

            Bu[0::nvar,1] = SpatialGradient[2,:]
            Bu[2::nvar,1] = -SpatialGradient[0,:]

            Bu[0::nvar,2] = -SpatialGradient[1,:]
            Bu[1::nvar,2] = SpatialGradient[0,:]

        elif SpatialGradient.shape[0]==2:
            # VORTICITY TERMS
            Bu[0::nvar,0] = -SpatialGradient[1,:]
            Bu[1::nvar,0] = SpatialGradient[0,:]


        BDB = 0.25*np.dot(Bu,Bu.T)
        return BDB


    def K_ww_Penalty_Integrand(self, Nw, Bases_w, ElectricDisplacementx,
        CoupleStressVector, kappa, analysis_nature="nonlinear", has_prestress=True):

        ndim = self.ndim
        nvar = self.nvar

        for ivar in range(ndim):
            Nw[ivar,ivar::nvar] = Bases_w

        NDN = kappa*np.dot(Nw.T,Nw)
        t=[]
        if analysis_nature == 'nonlinear' or has_prestress:
            t = np.dot(Nw,CoupleStressVector)
        return NDN, t



    def TractionIntegrand(self, B, SpatialGradient, ElectricDisplacementx,
        CauchyStressTensor, analysis_nature="nonlinear", has_prestress=True):
        """Applies to displacement potential based formulation"""
        pass



    def GetEnergy(self, function_space, material, LagrangeElemCoords,
        EulerELemCoords, ElectricPotentialElem, fem_solver, elem=0):
        """Get virtual energy of the system. For dynamic analysis this is handy for computing conservation of energy.
            The routine computes the global form of virtual internal energy i.e. integral of "W(C,G,C)"". This can be
            computed purely in a Lagrangian configuration.
        """

        nvar = self.nvar
        ndim = self.ndim
        nodeperelem = function_space.Bases.shape[0]

        det = np.linalg.det
        inv = np.linalg.inv
        Jm = function_space.Jm
        AllGauss = function_space.AllGauss

        internal_energy = 0.

        # COMPUTE KINEMATIC MEASURES AT ALL INTEGRATION POINTS USING EINSUM (AVOIDING THE FOR LOOP)
        # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
        ParentGradientX = np.einsum('ijk,jl->kil', Jm, LagrangeElemCoords)
        # MATERIAL GRADIENT TENSOR IN PHYSICAL ELEMENT [\nabla_0 (N)]
        MaterialGradient = np.einsum('ijk,kli->ijl', inv(ParentGradientX), Jm)
        # DEFORMATION GRADIENT TENSOR [\vec{x} \otimes \nabla_0 (N)]
        F = np.einsum('ij,kli->kjl', EulerELemCoords, MaterialGradient)

        # COMPUTE REMAINING KINEMATIC MEASURES
        StrainTensors = KinematicMeasures(F, fem_solver.analysis_nature)

        # SPATIAL GRADIENT AND MATERIAL GRADIENT TENSORS ARE EQUAL
        SpatialGradient = np.einsum('ikj',MaterialGradient)
        # COMPUTE ONCE detJ
        detJ = np.einsum('i,i->i',AllGauss[:,0],np.abs(det(ParentGradientX)))

        # GET ELECTRIC FIELD
        ElectricFieldx = - np.einsum('ijk,j',SpatialGradient,ElectricPotentialElem)

        # LOOP OVER GAUSS POINTS
        for counter in range(AllGauss.shape[0]):
            if material.energy_type == "enthalpy":
                # COMPUTE THE INTERNAL ENERGY AT THIS GAUSS POINT
                energy = material.InternalEnergy(StrainTensors,ElectricFieldx[counter,:],0,elem,counter)
            elif material.energy_type == "internal_energy":
                # COMPUTE ELECTRIC DISPLACEMENT IMPLICITLY
                ElectricDisplacementx = material.ElectricDisplacementx(StrainTensors, ElectricFieldx[counter,:], elem, counter)
                # COMPUTE THE INTERNAL ENERGY AT THIS GAUSS POINT
                energy = material.InternalEnergy(StrainTensors,ElectricDisplacementx[counter,:],elem,counter)

            # INTEGRATE INTERNAL ENERGY
            internal_energy += energy*detJ[counter]

        return internal_energy



    def Assemble(self, fem_solver, material, Eulerx, Eulerw, Eulers, Eulerp):

        # GET MESH DETAILS
        # C = mesh.InferPolynomialDegree() - 1
        formulation = self
        meshes = formulation.meshes
        mesh = meshes[0]
        nvar = formulation.nvar
        ndim = formulation.ndim
        nelem = meshes[0].nelem
        nodeperelem = meshes[0].elements.shape[1]
        local_size = int(ndim*meshes[0].elements.shape[1] + ndim*meshes[1].elements.shape[1] + ndim*meshes[2].elements.shape[1])
        capacity = local_size**2

        # ALLOCATE VECTORS FOR SPARSE ASSEMBLY OF STIFFNESS MATRIX - CHANGE TYPES TO INT64 FOR DoF > 1e09
        I_stiffness=np.zeros(int(capacity*nelem),dtype=np.int32)
        J_stiffness=np.zeros(int(capacity*nelem),dtype=np.int32)
        V_stiffness=np.zeros(int(capacity*nelem),dtype=np.float64)

        I_mass=[]; J_mass=[]; V_mass=[]
        if fem_solver.analysis_type !='static':
            # ALLOCATE VECTORS FOR SPARSE ASSEMBLY OF MASS MATRIX - CHANGE TYPES TO INT64 FOR DoF > 1e09
            I_mass=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
            J_mass=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
            V_mass=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.float64)

        # T = np.zeros((local_size,1),np.float64)
        T = np.zeros((mesh.points.shape[0]*nvar,1),np.float64)

        mass, F = [], []
        if fem_solver.has_moving_boundary:
            F = np.zeros((mesh.points.shape[0]*nvar,1),np.float64)


        if fem_solver.parallel:
            # COMPUATE ALL LOCAL ELEMENTAL MATRICES (STIFFNESS, MASS, INTERNAL & EXTERNAL TRACTION FORCES )
            # ParallelTuple = parmap.map(formulation.GetElementalMatrices,np.arange(0,nelem,dtype=np.int32),
                # function_space, mesh, material, fem_solver, Eulerx, Eulerp)

            ParallelTuple = parmap.map(formulation,np.arange(0,nelem,dtype=np.int32),
                function_space, mesh, material, fem_solver, Eulerx, Eulerp, processes= int(multiprocessing.cpu_count()/2))


        for elem in range(nelem):

            if fem_solver.parallel:
                # UNPACK PARALLEL TUPLE VALUES
                I_stiff_elem = ParallelTuple[elem][0]; J_stiff_elem = ParallelTuple[elem][1]; V_stiff_elem = ParallelTuple[elem][2]
                t = ParallelTuple[elem][3]; f = ParallelTuple[elem][4]
                I_mass_elem = ParallelTuple[elem][5]; J_mass_elem = ParallelTuple[elem][6]; V_mass_elem = ParallelTuple[elem][6]

            else:
                # COMPUATE ALL LOCAL ELEMENTAL MATRICES (STIFFNESS, MASS, INTERNAL & EXTERNAL TRACTION FORCES )
                I_stiff_elem, J_stiff_elem, V_stiff_elem, t, f, \
                I_mass_elem, J_mass_elem, V_mass_elem = formulation.GetElementalMatrices(elem,
                    formulation.function_spaces, formulation.meshes, material, fem_solver, Eulerx, Eulerw, Eulers, Eulerp)

            # SPARSE ASSEMBLY - STIFFNESS MATRIX
            SparseAssemblyNative(I_stiff_elem,J_stiff_elem,V_stiff_elem,I_stiffness,J_stiffness,V_stiffness,
                elem,nvar,nodeperelem,mesh.elements)

            if fem_solver.analysis_type != 'static' and fem_solver.is_mass_computed==False:
                # SPARSE ASSEMBLY - MASS MATRIX
                SparseAssemblyNative(I_mass_elem,J_mass_elem,V_mass_elem,I_mass,J_mass,V_mass,
                    elem,nvar,nodeperelem,mesh.elements)

            if fem_solver.has_moving_boundary:
                # RHS ASSEMBLY
                RHSAssemblyNative(F,f,elem,nvar,nodeperelem,mesh.elements)

            # INTERNAL TRACTION FORCE ASSEMBLY
            RHSAssemblyNative(T,t,elem,nvar,nodeperelem,mesh.elements)

            if (elem % fem_solver.assembly_print_counter == 0 or elem==nelem-1) and elem != 0:
                nume = elem+1 if elem==nelem-1 else elem
                print(('Assembled {} element matrices').format(nume))


        if fem_solver.parallel:
            del ParallelTuple
            gc.collect()


        stiffness = coo_matrix((V_stiffness,(I_stiffness,J_stiffness)),
            shape=((nvar*mesh.points.shape[0],nvar*mesh.points.shape[0])),dtype=np.float64).tocsr()

        # GET STORAGE/MEMORY DETAILS
        fem_solver.spmat = stiffness.data.nbytes/1024./1024.
        fem_solver.ijv = (I_stiffness.nbytes + J_stiffness.nbytes + V_stiffness.nbytes)/1024./1024.

        del I_stiffness, J_stiffness, V_stiffness
        gc.collect()

        if fem_solver.analysis_type != 'static' and fem_solver.is_mass_computed==False:
            mass = csr_matrix((V_mass,(I_mass,J_mass)),shape=((nvar*mesh.points.shape[0],
                nvar*mesh.points.shape[0])),dtype=np.float64)

            fem_solver.is_mass_computed = True

        return stiffness, T, F, mass



    def GetAugmentedSolution(self, fem_solver, material, TotalDisp, Eulerx, Eulerw, Eulers, Eulerp):
        """Get condensed variables
        """

        if self.save_condensed_matrices is False:
            return 0., 0.

        mesh = self.meshes[0]
        elements = mesh.elements
        points = mesh.points
        nelem = mesh.nelem
        nodeperelem = mesh.elements.shape[1]

        C = mesh.InferPolynomialDegree() - 1
        ndim = mesh.InferSpatialDimension()

        function_space = FunctionSpace(mesh, p=C+1, evaluate_at_nodes=True)

        Jm = function_space.Jm
        AllGauss = function_space.AllGauss

        AllEulerW = np.zeros((nelem,self.meshes[1].elements.shape[1],ndim))
        AllEulerS = np.zeros((nelem,self.meshes[2].elements.shape[1],ndim))

        NodalEulerW = np.zeros((self.meshes[1].points.shape[0],self.ndim))
        NodalEulerS = np.zeros((self.meshes[2].points.shape[0],self.ndim))

        # LOOP OVER ELEMENTS
        for elem in range(nelem):
            # GET THE FIELDS AT THE ELEMENT LEVEL
            LagrangeElemCoords = points[elements[elem,:],:]
            EulerELemCoords = Eulerx[elements[elem,:],:]

            if self.subtype == "lagrange_multiplier":
                k_uu = self.condensed_matrices['k_uu'][elem]
                k_us = self.condensed_matrices['k_us'][elem]
                k_ww = self.condensed_matrices['k_ww'][elem]
                k_ws = self.condensed_matrices['k_ws'][elem]
                inv_k_ws = self.condensed_matrices['inv_k_ws'][elem]
                tu = self.condensed_vectors['tu'][elem]
                tw = self.condensed_vectors['tw'][elem]
                ts = self.condensed_vectors['ts'][elem]

                EulerElemW = np.dot(inv_k_ws,(ts - np.dot(k_us.T,EulerELemCoords.ravel())[:,None])).ravel()
                EulerElemS = np.dot(inv_k_ws,(tw - np.dot(k_ww,EulerElemW)[:,None])).ravel()

                # SAVE
                AllEulerW[elem,:,:] = EulerElemW.reshape(self.meshes[1].elements.shape[1],ndim)
                AllEulerS[elem,:,:] = EulerElemW.reshape(self.meshes[2].elements.shape[1],ndim)


        for inode in self.all_nodes:
            Els, Pos = self.Elss[inode], self.Poss[inode]
            ncommon_nodes = Els.shape[0]
            for uelem in range(ncommon_nodes):
                NodalEulerW += AllEulerW[Els[uelem],Pos[uelem],:]
                NodalEulerS += AllEulerS[Els[uelem],Pos[uelem],:]
            # AVERAGE OUT
            NodalEulerW[inode,:] /= ncommon_nodes
            NodalEulerS[inode,:] /= ncommon_nodes

        # NAKE SURE TO UPDATE THESE INSTEAD OF CREATING THEM IN WHICH CASE YOU HAVE TO RETURN THEM
        Eulerw[:,:] += NodalEulerW
        Eulers[:,:] += NodalEulerS


        return NodalEulerW, NodalEulerS




import numpy as np
from numpy import einsum
from .MaterialBase import Material
from Florence.Tensor import trace, Voigt, UnVoigt

__all__ = ["IsotropicLinearFlexoelectricModel"]

class IsotropicLinearFlexoelectricModel(Material):
    """The couple stress model based on linear strain and curvature tensors

        W_cs(e,x) = W_lin(e) - E.P:e - 1/2. eps e.e + eta*x:x

    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(IsotropicLinearFlexoelectricModel, self).__init__(mtype, ndim, **kwargs)
        # REQUIRES SEPARATELY
        self.energy_type = "enthalpy"
        self.nature = "linear"
        self.fields = "flexoelectric"

        # FOR STATICALLY CONDENSED FORMULATION
        if self.ndim==3:
            self.elasticity_tensor_size = 9
            self.H_VoigtSize = 9
        elif self.ndim==2:
            self.elasticity_tensor_size = 5
            self.H_VoigtSize = 5
        self.gradient_elasticity_tensor_size = ndim

        # LOW LEVEL DISPATCHER
        # self.has_low_level_dispatcher = True
        self.has_low_level_dispatcher = False

    def KineticMeasures(self,F,ElectricFieldx=0, elem=0):
        D = self.eps*ElectricFieldx.reshape(ElectricFieldx.shape[0],ElectricFieldx.shape[1],1) # CHECK
        sigma = np.zeros((ElectricFieldx.shape[0],self.ndim,self.ndim))
        return (D, sigma, self.H_Voigt, self.elasticity_tensors, self.gradient_elasticity_tensors,
            self.piezoelectric_tensors, self.flexoelectric_tensors, self.dielectric_tensors)


    def Hessian(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        eta = self.eta
        eps = self.eps
        I = StrainTensors['I']

        self.elasticity_tensor = Voigt(lamb*einsum('ij,kl',I,I)+mu*(einsum('ik,jl',I,I)+einsum('il,jk',I,I)),1)
        self.gradient_elasticity_tensor = 2.*eta*I
        self.coupling_tensor0 = np.zeros((self.elasticity_tensor.shape[0],self.gradient_elasticity_tensor.shape[0]))

        # Piezoelectric tensor must be 6x3 [3D] or 3x2 [2D]
        self.piezoelectric_tensor = self.P
        # Dielectric tensor
        self.dielectric_tensor = -self.eps*I

        # # BUILD HESSIAN
        factor = -1.
        H1 = np.concatenate((self.elasticity_tensor,factor*self.piezoelectric_tensor),axis=1)
        H2 = np.concatenate((factor*self.piezoelectric_tensor.T,self.dielectric_tensor),axis=1)
        H_Voigt = np.concatenate((H1,H2),axis=0)
        return H_Voigt



    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        eta = self.eta
        piezoelectric_tensor = self.P
        flexoelectric_tensor = self.f
        I = StrainTensors['I']
        e = StrainTensors['strain'][gcounter]
        E = ElectricFieldx.reshape(self.ndim,1)

        tre = trace(e)
        if self.ndim==2:
            tre +=1.


        # REQUIRES CHECKS
        if self.ndim == 3:
            sigma_f = np.dot(flexoelectric_tensor,E)
            sigma_f = np.array([
                [0.,-sigma_f[2],sigma_f[1]],
                [sigma_f[2],0.,-sigma_f[0]],
                [-sigma_f[1],sigma_f[0],0.],
                ])
        else:
            # THIRD COMPONENT WHICH DOES NOT EXITS?
            sigma_f = np.zeros((self.ndim,self.ndim))

        sigma = lamb*tre*I + 2.0*mu*e - UnVoigt(np.dot(piezoelectric_tensor,E)) + sigma_f
        return sigma


    def CoupleStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        eta = self.eta
        I = StrainTensors['I']
        F = StrainTensors['F'][gcounter]
        GradU = F - I
        k = GradU - GradU.T
        k = np.array([k[0,1],k[0,2],k[1,2]])

        couple_stress = 2*eta*k
        return couple_stress


    def LagrangeMultiplierStress(self,S,ElectricFieldx=None,elem=0,gcounter=0):

        lm_stress = 0.5*S
        return lm_stress


    def TotalStress(self,StrainTensors,S,ElectricFieldx=None,elem=0,gcounter=0):

        cauchy_stress = self.CauchyStress(StrainTensors,ElectricFieldx,elem=elem,gcounter=gcounter)
        lm_stress     = self.LagrangeMultiplierStress(S,ElectricFieldx,elem=elem,gcounter=gcounter)
        lm_stress     = np.array([
            [0.,lm_stress[0],lm_stress[1]],
            [-lm_stress[0],0.,lm_stress[2]],
            [-lm_stress[1],-lm_stress[2],0.]
            ])
        return cauchy_stress + lm_stress


    def ElectricDisplacementx(self,StrainTensors,ElectricFieldx,EulerW=0,elem=0,gcounter=0):
        # CHECK
        piezoelectric_tensor = self.P
        flexoelectric_tensor = self.f
        return self.eps*ElectricFieldx.reshape(self.ndim,1) +  np.dot(flexoelectric_tensor.T,EulerW)


    def InternalEnergy(self,StrainTensors,ElectricFieldx,EulerW=0,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        eta = self.eta
        eps = self.eps

        e = StrainTensors['strain'][gcounter]
        E = ElectricFieldx.reshape(self.ndim)

        tre = trace(e)
        if self.ndim==2:
            tre +=1.

        self.strain_energy = mu*einsum("ij,ij",e,e) + lamb*tre**2
        self.electrical_energy = eps*mu*einsum("i,i",E,E)
        return self.strain_energy, self.electrical_energy



import numpy as np
from Florence import QuadratureRule, FunctionSpace, Mesh
from Florence.FiniteElements.LocalAssembly._KinematicMeasures_ import _KinematicMeasures_
from Florence.VariationalPrinciple._GeometricStiffness_ import GeometricStiffnessIntegrand as GetGeomStiffness
from .DisplacementApproachIndices import FillGeometricB
from ._MassIntegrand_ import __MassIntegrand__, __ConstantMassIntegrand__


__all__ = ["VariationalPrinciple"]


class VariationalPrinciple(object):

    energy_dissipation = []
    internal_energy = []
    kinetic_energy = []
    external_energy = []

    power_dissipation = []
    internal_power = []
    kinetic_power = []
    external_power = []


    def __init__(self, mesh, variables_order=(1,0),
        analysis_type='static', analysis_nature='nonlinear', fields='mechanics',
        quadrature_rules=None, median=None, quadrature_type=None,
        function_spaces=None, compute_post_quadrature=True):

        self.variables_order = variables_order
        self.nvar = None
        self.ndim = mesh.points.shape[1]

        if isinstance(self.variables_order,int):
            self.variables_order = tuple(self.variables_order)

        self.quadrature_rules = quadrature_rules
        self.quadrature_type = quadrature_type
        self.function_spaces = function_spaces
        self.median = median
        self.analysis_type = analysis_type
        self.analysis_nature = analysis_nature
        self.fields = fields

        self.compute_post_quadrature = compute_post_quadrature


        # GET NUMBER OF VARIABLES
        self.GetNumberOfVariables()

        # # BUILD MESHES FOR THESE ORDRES
        # p = mesh.InferPolynomialDegree()

        # if p != self.variables_order[0] and self.variables_order[0]!=0:
        #     mesh.GetHighOrderMesh(C=self.variables_order[0]-1)

        # if len(self.variables_order) == 2:
        #     if self.variables_order[2] == 0 or self.variables_order[1]==0 or self.variables_order[0]==0:
        #         # GET MEDIAN OF THE ELEMENTS FOR CONSTANT VARIABLES
        #         self.median, self.bases_at_median = mesh.Median

        #     # GET QUADRATURE RULES
        #     self.quadrature_rules = [None]*len(self.variables_order)
        #     QuadratureOpt = 3
        #     for counter, degree in enumerate(self.variables_order):
        #         if degree==0:
        #             degree=1
        #         # self.quadrature_rules[counter] = list(GetBasesAtInegrationPoints(degree-1, 2*degree,
        #             # QuadratureOpt,mesh.element_type))
        #         quadrature = QuadratureRule(optimal=QuadratureOpt,
        #             norder=2*degree, mesh_type=mesh.element_type)
        #         self.quadrature_rules[counter] = quadrature
        #         # FunctionSpace(mesh, p=degree, 2*degree, QuadratureOpt,mesh.element_type)
        #         # self.quadrature_rules[counter] = list()


    def GetQuadratureOrder(self,C,element_type):
        """Finds quadrature degree/strength for a given polynomial order C=p-1 [where is polynomial degree]"""
        if element_type == "tri" or element_type == "tet":
            norder = 2*C if C > 0 else 1
            norder_post = 2*(C+1)
        else:
            norder = C+2
            norder_post = 2*(C+2)

        return norder, norder_post


    def GetVolume(self, function_space, LagrangeElemCoords, EulerELemCoords, requires_geometry_update, elem=0):
        """ Find the volume (area in 2D) of element [could be curved or straight]
        """

        # GET LOCAL KINEMATICS
        detJ = _KinematicMeasures_(function_space.Jm, function_space.AllGauss[:,0],
            LagrangeElemCoords, EulerELemCoords, requires_geometry_update)[2]

        # volume = 0.
        # LOOP OVER GAUSS POINTS
        # for counter in range(function_space.AllGauss.shape[0]):
            # volume += detJ[counter]

        # return volume

        return detJ.sum()


    def ConstitutiveStiffnessIntegrand(self, B, SpatialGradient, CauchyStressTensor, H_Voigt,
        analysis_nature="nonlinear", has_prestress=True):
        """Applies to displacement based formulation"""
        pass

    def __ConstitutiveStiffnessIntegrand__(self, SpatialGradient, H_Voigt, CauchyStressTensor, detJ,
        analysis_nature="nonlinear", has_prestress=False, requires_geometry_update=True):
        """Applies to displacement based formulation"""
        pass


    def GeometricStiffnessIntegrand(self, SpatialGradient, CauchyStressTensor):
        """Applies to displacement based, displacement potential based and all mixed
        formulations that involve static condensation"""

        ndim = self.ndim
        nvar = self.nvar

        B = np.zeros((nvar*SpatialGradient.shape[0],ndim*ndim))
        S = np.zeros((ndim*ndim,ndim*ndim))
        SpatialGradient = SpatialGradient.T.copy('c')

        FillGeometricB(B,SpatialGradient,S,CauchyStressTensor,ndim,nvar)

        BDB = np.dot(np.dot(B,S),B.T)

        return BDB


    def __GeometricStiffnessIntegrand__(self, SpatialGradient, CauchyStressTensor, detJ):
        """Applies to displacement based formulation"""
        return GetGeomStiffness(np.ascontiguousarray(SpatialGradient),CauchyStressTensor, detJ, self.nvar)


    # def TractionIntegrand(self, B, SpatialGradient, CauchyStressTensor,
    #     analysis_nature="nonlinear", has_prestress=True):
    #     """Applies to displacement based formulation"""

    #     SpatialGradient = SpatialGradient.T.copy()
    #     t=[]
    #     if analysis_nature == 'nonlinear' or has_prestress:
    #         TotalTraction = GetTotalTraction(CauchyStressTensor)
    #         t = np.dot(B,TotalTraction)

    #     return t


    def TractionIntegrand(self, B, SpatialGradient, CauchyStressTensor,
        analysis_nature="nonlinear", has_prestress=True):
        """Applies to displacement based formulation"""
        pass


    def __TractionIntegrand__(self, B, SpatialGradient, CauchyStressTensor,
        analysis_nature="nonlinear", has_prestress=True):
        """Applies to displacement based formulation"""
        pass


    def MassIntegrand(self, Bases, N, material):

        nvar = self.nvar
        ndim = self.ndim
        rho = material.rho

        for ivar in range(ndim):
            N[ivar::nvar,ivar] = Bases

        rhoNN = rho*np.dot(N,N.T)
        return rhoNN

    def GetConstantMassIntegrand(self, Domain, material):
        # COMPUTE THE CONSTANT PART OF MASS MATRIX 'rho*np.dot(N,N.T)'
        # THIS SHOULD BE CALLED ONCE AS IT IS THE SAME FOR EVERY ELEMENT
        # N = np.zeros((Domain.Bases.shape[0]*self.nvar,self.nvar))
        # rhoNN_all = np.zeros((Domain.Bases.shape[0]*self.nvar,Domain.Bases.shape[0]*self.nvar,Domain.AllGauss.shape[0]))
        # # LOOP OVER GAUSS POINTS
        # for counter in range(0,Domain.AllGauss.shape[0]):
        #     # COMPUTE THE MASS INTEGRAND
        #     rhoNN_all[:,:,counter] = self.MassIntegrand(Domain.Bases[:,counter],N,material)

        # self.constant_mass_integrand = rhoNN_all
        # return rhoNN_all

        N = np.zeros((Domain.Bases.shape[0]*self.nvar,self.nvar))
        rhoNN_all = np.zeros((Domain.AllGauss.shape[0],Domain.Bases.shape[0]*self.nvar,Domain.Bases.shape[0]*self.nvar))
        # LOOP OVER GAUSS POINTS
        for counter in range(0,Domain.AllGauss.shape[0]):
            # COMPUTE THE MASS INTEGRAND
            rhoNN_all[counter,:,:] = self.MassIntegrand(Domain.Bases[:,counter],N,material)

        self.constant_mass_integrand = rhoNN_all
        return rhoNN_all


    def GetLocalMass(self, function_space, material, LagrangeElemCoords, EulerELemCoords, fem_solver, elem):

        ndim = self.ndim
        nvar = self.nvar
        Domain = function_space

        N = np.zeros((Domain.Bases.shape[0]*nvar,nvar))
        mass = np.zeros((Domain.Bases.shape[0]*nvar,Domain.Bases.shape[0]*nvar))

        # LOOP OVER GAUSS POINTS
        for counter in range(0,Domain.AllGauss.shape[0]):
            # GRADIENT TENSOR IN PARENT ELEMENT [\nabla_\varepsilon (N)]
            Jm = Domain.Jm[:,:,counter]
            Bases = Domain.Bases[:,counter]
            # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
            ParentGradientX=np.dot(Jm,LagrangeElemCoords)

            # COMPUTE THE MASS INTEGRAND
            rhoNN = self.MassIntegrand(Bases,N,material)
            # INTEGRATE MASS
            mass += rhoNN*Domain.AllGauss[counter,0]*np.abs(np.linalg.det(ParentGradientX))

        return mass

    def __GetLocalMass__(self, function_space, material, LagrangeElemCoords, EulerELemCoords, fem_solver, elem):
        # GET LOCAL KINEMATICS
        _, _, detJ = _KinematicMeasures_(function_space.Jm, function_space.AllGauss[:,0], LagrangeElemCoords,
            EulerELemCoords, False)
        return __MassIntegrand__(material.rho,function_space.Bases,detJ,material.ndim,material.nvar)


    def GetLocalMass_Efficient(self, function_space, material, LagrangeElemCoords, EulerELemCoords, fem_solver, elem):
        """Computes elemental mass matrix in a very efficient way by computing rhoNN only once"""

        Domain = function_space
        # MAPPING TENSOR [\partial\vec{X}/ \partial\vec{\varepsilon} (ndim x ndim)]
        ParentGradientX = np.einsum('ijk,jl->kil', Domain.Jm, LagrangeElemCoords)
        # COMPUTE ONCE detJ
        detJ = np.einsum('i,i->i',Domain.AllGauss[:,0],np.abs(np.linalg.det(ParentGradientX)))
        # INTEGRATE MASS
        # mass = np.einsum("ijk,k",self.constant_mass_integrand,detJ)
        return np.einsum("ijk,i",self.constant_mass_integrand,detJ)

        # Domain = function_space
        # rhoNN_all = self.constant_mass_integrand
        # mass = np.zeros((Domain.Bases.shape[0]*self.nvar,Domain.Bases.shape[0]*self.nvar))
        # LOOP OVER GAUSS POINTS
        # for counter in range(0,Domain.AllGauss.shape[0]):
        #     # INTEGRATE MASS
        #     mass += rhoNN_all[:,:,counter]*detJ[counter]
        # return mass

    def __GetLocalMass_Efficient__(self, function_space, material, LagrangeElemCoords, EulerELemCoords, fem_solver, elem):
        """Computes elemental mass matrix in a very efficient way by computing rhoNN only once"""
        # GET LOCAL KINEMATICS
        _, _, detJ = _KinematicMeasures_(function_space.Jm, function_space.AllGauss[:,0], LagrangeElemCoords,
            EulerELemCoords, False)
        return __ConstantMassIntegrand__(self.constant_mass_integrand,detJ)

    def GetLumpedMass(self,mass):
        return np.sum(mass,1)[:,None]

    def VerifyMass(self,density, ndim, mass, detJ):
        if np.isclose(mass.sum(),ndim*density*detJ.sum()):
            raise RuntimeError("Local mass matrix is not computed correctly")
        return


    def FindIndices(self,A):
        return self.local_rows, self.local_columns, A.ravel()
        # return self.local_rows, self.local_columns, A.flatten()



    def GetNumberOfVariables(self):
        """Returns (self.nvar) i.e. number of variables/unknowns per node, for the formulation.
            Note that self.nvar does not take into account the unknowns which get condensated
        """

        # nvar = 0
        # for i in self.variables_order:
        #     # DO NOT COUNT VARIABLES THAT GET CONDENSED OUT
        #     if i!=0:
        #         if mesh.element_type == "tri":
        #             nvar += (i+1)*(i+2) // 2
        #         elif mesh.element_type == "tet":
        #             nvar += (i+1)*(i+2)*(i+3) // 6
        #         elif mesh.element_type == "quad":
        #             nvar += (i+1)**2
        #         elif mesh.element_type == "hex":
        #             nvar += (i+1)**3

        # nvar = sum(self.variables_order)
        if self.nvar == None:
            self.nvar = self.ndim
        return self.nvar


    def __call__(self,*args,**kwargs):
        """This is purely to get around pickling while parallelising"""
        return self.GetElementalMatrices(*args,**kwargs)
        # return self.GetElementalMatricesInVectorForm(*args,**kwargs)

























class upFormulation():
    pass


class FiveFieldPenalty(VariationalPrinciple):

    def __init__(self,mesh):

        self.submeshes = []*4
        # PREPARE SUBMESHES
        if mesh.element_type == "tri":
            self.submeshes[0].elements = mesh.elements[:,:3]
            self.submeshes[0].nelem = mesh.elements.shape[0]
            self.submeshes[0].points = mesh.points[:self.submeshes[0].elements.max()+1,:]
            self.submeshes[0].element_type = "tri"







#cython: profile=False
#cython: infer_types=True
#cython: boundscheck=False
#cython: wraparound=False
#cython: nonecheck=False

import numpy as np
cimport numpy as np

ctypedef double Real


cdef extern from "_GeometricStiffness_.h":
    inline void _GeometricStiffnessFiller_(Real *geometric_stiffness, 
        const Real *SpatialGradient, const Real *CauchyStressTensor, const Real *detJ, 
        const int ndim, const int nvar, const int nodeperelem, const int nguass) nogil

def GeometricStiffnessIntegrand(np.ndarray[Real, ndim=3, mode='c'] SpatialGradient,
    np.ndarray[Real, ndim=3, mode='c'] CauchyStressTensor,
    np.ndarray[Real, ndim=1] detJ, int nvar):

    cdef int ngauss = SpatialGradient.shape[0]
    cdef int nodeperelem = SpatialGradient.shape[1]
    cdef int ndim = SpatialGradient.shape[2]
    cdef int local_size = nvar*nodeperelem

    cdef np.ndarray[Real, ndim=2, mode='c'] geometric_stiffness = np.zeros((local_size,
        local_size),dtype=np.float64)

    _GeometricStiffnessFiller_(&geometric_stiffness[0,0], &SpatialGradient[0,0,0], 
        &CauchyStressTensor[0,0,0], &detJ[0], ndim, nvar, nodeperelem, ngauss)


    return geometric_stiffness


import numpy as np
from .MaterialBase import Material
from Florence.Tensor import trace


class LinearElastic(Material):
    """Classical linear elastic material model

        W = lamb/2.*trace(e)**2 + mu*e*e

    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(LinearElastic, self).__init__(mtype, ndim, **kwargs)
        self.energy_type = "internal_energy"
        self.nature = "linear"
        self.fields = "mechanics"

        if self.ndim==3:
            self.H_VoigtSize = 6
        elif self.ndim==2:
            self.H_VoigtSize = 3

        # LOW LEVEL DISPATCHER
        # self.has_low_level_dispatcher = False
        self.has_low_level_dispatcher = True

    def KineticMeasures(self, F, ElectricFieldx=0, elem=0):
        from Florence.MaterialLibrary.LLDispatch._LinearElastic_ import KineticMeasures
        # STRAIN TENSOR IS COMPUTED WITHIN THE LOW LEVEL VERSION OF THE MATERIAL
        return KineticMeasures(self,F)


    def Hessian(self,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):
        return self.H_Voigt


    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        strain = StrainTensors['strain'][gcounter]
        I = StrainTensors['I']

        mu = self.mu
        lamb = self.lamb

        tre = trace(strain)
        if self.ndim == 2:
            tre += 1.

        return 2*mu*strain + lamb*tre*I


    def ElectricDisplacementx(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        return np.zeros((self.ndim,1))


    def InternalEnergy(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        strain = StrainTensors['strain'][gcounter]

        if self.ndim == 3:
            tre = trace(strain)
        elif self.ndim == 2:
            tre = trace(strain) + 1

        energy = lamb/2.*tre**2 + mu*np.einsum("ij,ij",strain,strain)
        return energy



import numpy as np
from numpy import einsum
from .MaterialBase import Material
from Florence.Tensor import trace, Voigt


class IsotropicElectroMechanics_1(Material):
    """docstring for IsotropicElectroMechanics"""

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(IsotropicElectroMechanics_1, self).__init__(mtype, ndim, **kwargs)

        self.nvar = self.ndim+1
        self.energy_type = "enthalpy"
        self.nature = "nonlinear"
        self.fields = "electro_mechanics"

        if self.ndim == 2:
            self.H_VoigtSize = 5
        elif self.ndim == 3:
            self.H_VoigtSize = 9

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = False


    def Hessian(self,StrainTensors, ElectricFieldx=0, elem=0, gcounter=0):

        mu = self.mu
        lamb = self.lamb
        varepsilon_1 = self.eps_1

        detF = StrainTensors['J'][gcounter]


        mu2 = mu - lamb*(detF-1.0)
        lamb2 = lamb*(2.0*detF-1.0) - mu

        delta = StrainTensors['I']
        E = 1.0*ElectricFieldx

        Ex = E.reshape(E.shape[0])
        EE = np.outer(E,E)
        innerEE = np.dot(E,E.T)

        I = delta
        # C = lamb2*AijBkl(I,I) +mu2*(AikBjl(I,I)+AilBjk(I,I)) + varepsilon_1*(AijBkl(I,EE) + AijBkl(EE,I) - \
        # 2.*AikBjl(EE,I)-2.0*AilBjk(I,EE) ) + varepsilon_1*(np.dot(E.T,E)[0,0])*(AikBjl(I,I)-0.5*AijBkl(I,I))

        # ORIGINAL
        # C = lamb2*AijBkl(I,I) +mu2*(AikBjl(I,I)+AilBjk(I,I)) +\
        #     varepsilon_1*(AijBkl(I,EE) + AijBkl(EE,I) -AikBjl(EE,I)-AilBjk(EE,I)-AilBjk(I,EE)-AikBjl(I,EE) ) +\
        #     varepsilon_1*(np.dot(E.T,E)[0,0])*(0.5*(AikBjl(I,I) + AilBjk(I,I))-0.5*AijBkl(I,I))
        # C=0.5*(C+C.T)
        # C_Voigt = C

        C = lamb2*einsum("ij,kl",I,I) +mu2*(einsum("ik,jl",I,I)+einsum("il,jk",I,I)) +\
            varepsilon_1*(einsum("ij,kl",I,EE) + einsum("ij,kl",EE,I) - einsum("ik,jl",EE,I)- einsum("il,jk",I,EE) -\
            einsum("il,jl",I,EE)- einsum("ik,jl",I,EE) ) +\
            varepsilon_1*(innerEE)*(0.5*( einsum("ik,jl",I,I)+einsum("il,jk",I,I) )-0.5* einsum("ij,kl",I,I) )
        C_Voigt = Voigt(C,1)


        # Computing the hessian
        # Elasticity tensor (C - 4th order tensor)
        # C[i,j,k,l] += lamb2*delta[i,j]*delta[k,l]+2.0*mu2*(delta[i,k]*delta[j,l]) #

        b = StrainTensors['b'][gcounter]
        be = np.dot(b,ElectricFieldx).reshape(self.ndim,1)
        # Coupled Tensor (e - 3rd order)

        # e[k,i,j] += (-2.0*varepsilon_1/detF)*(be[j]*b[i,k] + be[i]*b[j,k]) #
        # e[i,j,k] += 1.0*varepsilon_1*( E[i]*delta[j,k] + E[j]*delta[i,k] - delta[i,j]*E[k]) ##
        # e[k,i,j] += 1.0*varepsilon_1*(E[i]*delta[j,k] + E[j]*delta[i,k] - delta[i,j]*E[k]) ##

        # Note that the actual piezoelectric tensor is symmetric wrt to the last two indices
        # Actual tensor is: e[k,i,j] += 1.0*varepsilon_1*(E[i]*delta[j,k] + E[j]*delta[i,k] - delta[i,j]*E[k])
        # We need to make its Voigt_form symmetric with respect to (j,k) instead of (i,j)

        # ORIGINAL
        # e_voigt = 1.0*varepsilon_1*(AijUk(I,Ex)+AikUj(I,Ex)-UiAjk(Ex,I)).T

        e_voigt = 1.0*varepsilon_1*( einsum('ij,k',I,Ex) + einsum('ik,j',I,Ex) - einsum('i,jk',Ex,I) ).T
        e_voigt = Voigt(np.ascontiguousarray(e_voigt),1)

        # Dielectric Tensor (Permittivity - 2nd order)
        Permittivity = -varepsilon_1*delta ##

        # bb =  np.dot(StrainTensors.b,StrainTensors.b) #
        # Permittivity = -(2.0*varepsilon_1/detF)*bb #


        factor = -1.
        H1 = np.concatenate((C_Voigt,factor*e_voigt),axis=1)
        H2 = np.concatenate((factor*e_voigt.T,Permittivity),axis=1)
        H_Voigt = np.concatenate((H1,H2),axis=0)

        self.H_VoigtSize = H_Voigt.shape[0]


        # return H_Voigt, C, e, Permittivity
        return H_Voigt


    def CauchyStress(self, StrainTensors, ElectricFieldx, elem=0,gcounter=0):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        E = ElectricFieldx

        mu = self.mu
        lamb = self.lamb
        varepsilon_1 = self.eps_1

        be = np.dot(b,ElectricFieldx)

        return 1.0*mu/J*b+(lamb*(J-1.0)-mu)*I + varepsilon_1*(np.dot(E,E.T)-0.5*np.dot(E.T,E)*I) ##
        # return 1.0*mu/J*b+(lamb*(J-1.0)-mu)*I - (2.0*varepsilon_1/J)*np.dot(be,be.T)


    def ElectricDisplacementx(self, StrainTensors, ElectricFieldx, elem=0, gcounter=0):

        varepsilon_1 = self.eps_1
        return varepsilon_1*ElectricFieldx[:,None] ##

        # J = StrainTensors['J'][gcounter]
        # b = StrainTensors['b'][gcounter]
        # bb =  np.dot(b,b)
        # return (2.0*varepsilon_1/StrainTensors.J)*np.dot(bb,ElectricFieldx).reshape(StrainTensors.b.shape[0],1)




from __future__ import print_function
import os, sys, imp
from time import time
import multiprocessing as MP
import numpy as np
# try:
#     from mpi4py import MPI
#     hasMPI = True
#     comm = MPI.COMM_WORLD
# except ImportError:
#     hasMPI = False





# FLORENCE BASE CLASS 
class Base(object):
    """FLorence base class. General data such as directories, files, analysis session, etc 
        that needs to be loaded a priori are stored
        
        pwd:                        Florence's top level directory
        session:                    {'FEM','BEM','Coupled'} Session to run
        __NO_DEBUG__:               Enter debug mode of the package (if false). Activates all numerical checks                 
        __VECTORISATION__:          Activate numpy's (einsum) for computing elemental matrices with no loops
        __PARALLEL__:               Activate multiprocessing for either shared or distributed memory or both
        numCPU:                     Number of concurrent cores/hyperthreads for parallelisation
        __MEMORY__:                 {'SHARED','DISTRIBUTED','AUTO','HYBRID'} Option for shared/distributed 
                                    memory parallelisation
        
        C:                          [int] order of basis functions. Note that C=P-1 where P is polynomial degree
        norder:                     [int] number of quadrature points 
        plot:                       [tuple of ints] plot flag for BEM
        nrplot:                     [tuple] plot flag for Newton-Raphson convergence
        write:                      [boolean] flag for writting simulation results in .vtu/.mat/.eps/.dat formats
        
        """

    FloatType=np.float32
    IntType=np.int32

    if sys.version_info.major == 2:
        Range = xrange
    else:
        Range = range 


    pwd = os.path.dirname(os.path.realpath('__file__'))
    session = 'FEM'
    # session = 'BEM'
    # session = 'Coupled'


    __NO_DEBUG__ = False
    __VECTORISATION__ = True
    __PARALLEL__ = False
    nCPU = 1
    __MEMORY__ = 'SHARED'


    C = 0
    norder = 2 
    plot = (0,3)
    nrplot = (0,'last')
    write = 0



    # # PROBLEM SPATIAL DIMENSION- 1D, 2D, 3D
    # ndim = 2
    # nvar = ndim
    # Fields = 'Mechanics'
    # # Fields = 'ElectroMechanics'
    
    # Formulation = 'DisplacementApproach'
    # # Formulation = 'DisplacementElectricPotentialApproach'

    # Analysis = 'Static'
    # # Analysis = 'Dynamic'
    # AnalysisType = 'Linear'
    # # AnalysisType = 'Nonlinear'


    Timer = 0


    # DECIDE WHICH PARALLEL MODEL TO ACTIVATE
    def ParallelModel(self):
        if self.__MEMORY__ == "shared":
            pass
        elif self.__MEMORY__ == "distributed":
            print(comm.rank)

    isScaledJacobianComputed = False



from __future__ import print_function
import gc, os, sys
import numpy as np
import scipy as sp
import numpy.linalg as la
import scipy.linalg as sla
from numpy.linalg import norm
from time import time
from copy import deepcopy
from warnings import warn
from time import time

from Florence.FiniteElements.Assembly import Assemble
from Florence import Mesh

__all__ = ["StructuralDynamicIntegrator"]


class StructuralDynamicIntegrator(object):
    """Base class for structural time integerators
    """

    def __init__(self):
        super(StructuralDynamicIntegrator, self).__init__()
        self.gamma   = 0.5
        self.beta    = 0.25
        self.lump_rhs = False


    def GetBoundaryInfo(self, mesh, formulation, boundary_condition):

        all_dofs = np.arange(mesh.points.shape[0]*formulation.nvar)
        if formulation.fields == "electro_mechanics":
            self.electric_dofs = all_dofs[formulation.nvar-1::formulation.nvar]
            self.mechanical_dofs = np.setdiff1d(all_dofs,self.electric_dofs)

            # GET BOUNDARY CONDITON FOR THE REDUCED MECHANICAL SYSTEM
            self.columns_in_mech = np.intersect1d(boundary_condition.columns_in,self.mechanical_dofs)
            self.columns_in_mech_idx = np.in1d(self.mechanical_dofs,boundary_condition.columns_in)

            # GET BOUNDARY CONDITON FOR THE REDUCED ELECTROSTATIC SYSTEM
            self.columns_in_electric = np.intersect1d(boundary_condition.columns_in,self.electric_dofs)
            self.columns_in_electric_idx = np.in1d(self.electric_dofs,boundary_condition.columns_in)


            # GET FREE MECHANICAL DOFs
            self.columns_out_mech = np.intersect1d(boundary_condition.columns_out,self.mechanical_dofs)
            self.columns_out_mech_idx = np.in1d(self.mechanical_dofs,boundary_condition.columns_out)
            self.columns_out_mech_reverse_idx = np.in1d(boundary_condition.columns_out,self.columns_out_mech)

            # GET FREE ELECTROSTATIC DOFs
            self.columns_out_electric = np.intersect1d(boundary_condition.columns_out,self.electric_dofs)
            self.columns_out_electric_idx = np.in1d(self.electric_dofs,boundary_condition.columns_out)
            self.columns_out_electric_reverse_idx = np.in1d(boundary_condition.columns_out,
                self.columns_out_electric)

            self.applied_dirichlet_mech = boundary_condition.applied_dirichlet[self.columns_out_mech_reverse_idx]
            self.applied_dirichlet_electric = boundary_condition.applied_dirichlet[self.columns_out_electric_reverse_idx]

            # MAPPED QUANTITIES
            out_idx = np.in1d(all_dofs,boundary_condition.columns_out)
            idx_electric = all_dofs[formulation.nvar-1::formulation.nvar]
            idx_mech = np.setdiff1d(all_dofs,idx_electric)

            self.all_electric_dofs = np.arange(mesh.points.shape[0])
            self.electric_out = self.all_electric_dofs[out_idx[idx_electric]]
            self.electric_in = np.setdiff1d(self.all_electric_dofs,self.electric_out)

            self.all_mech_dofs = np.arange(mesh.points.shape[0]*formulation.ndim)
            self.mech_out = self.all_mech_dofs[out_idx[idx_mech]]
            self.mech_in = np.setdiff1d(self.all_mech_dofs,self.mech_out)

        elif formulation.fields == "mechanics":
            self.electric_dofs = []
            self.mechanical_dofs = all_dofs
            self.columns_out_mech = boundary_condition.columns_out
            self.columns_out_mech_reverse_idx = np.ones_like(self.columns_out_mech).astype(bool)

            self.mech_in = boundary_condition.columns_in
            self.mech_out = boundary_condition.columns_out

            self.applied_dirichlet_mech = boundary_condition.applied_dirichlet




    def ComputeEnergyDissipation(self,function_space,mesh,material,formulation,fem_solver,
        Eulerx, TotalDisp, NeumannForces, M, velocities, Increment):

        internal_energy = 0.
        for elem in range(mesh.nelem):
            LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
            EulerElemCoords = Eulerx[mesh.elements[elem,:],:]

            internal_energy += formulation.GetEnergy(function_space, material,
                LagrangeElemCoords, EulerElemCoords, fem_solver, elem)

        if formulation.fields == "electro_mechanics":
            M_mech = M[self.mechanical_dofs,:][:,self.mechanical_dofs]
            kinetic_energy = 0.5*np.dot(velocities.ravel(),M_mech.dot(velocities.ravel()))
        else:
            kinetic_energy = 0.5*np.dot(velocities.ravel(),M.dot(velocities.ravel()))

        external_energy = np.dot(TotalDisp[:,:,Increment].ravel(),NeumannForces[:,Increment])

        total_energy = internal_energy + kinetic_energy - external_energy
        return total_energy, internal_energy, kinetic_energy, external_energy



    def ComputePowerDissipation(self,function_space,mesh,material,formulation,fem_solver, Eulerx, TotalDisp,
        NeumannForces, M, velocities, accelerations, Increment):

        internal_energy = 0.
        for elem in range(mesh.nelem):
            LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
            EulerElemCoords    = Eulerx[mesh.elements[elem,:],:]
            VelocityElem       = velocities[mesh.elements[elem,:],:]

            internal_energy += formulation.GetLinearMomentum(function_space, material,
                LagrangeElemCoords, EulerElemCoords, VelocityElem, fem_solver, elem)

        if formulation.fields == "electro_mechanics":
            M_mech = M[self.mechanical_dofs,:][:,self.mechanical_dofs]
            kinetic_energy = np.dot(velocities.ravel(),M_mech.dot(accelerations.ravel()))
        else:
            kinetic_energy = np.dot(velocities.ravel(),M.dot(accelerations.ravel()))

        external_energy = np.dot(velocities.ravel(),NeumannForces[:,Increment])

        total_energy = internal_energy + kinetic_energy - external_energy
        return total_energy, internal_energy, kinetic_energy, external_energy






    def UpdateFixMechanicalDoFs(self, AppliedDirichletInc, fsize, nvar):
        """Updates the geometry (DoFs) with incremental Dirichlet boundary conditions
            for fixed/constrained degrees of freedom only. Needs to be applied per time steps"""

        # GET TOTAL SOLUTION
        TotalSol = np.zeros((fsize,1))
        TotalSol[self.mech_out,0] = AppliedDirichletInc

        # RE-ORDER SOLUTION COMPONENTS
        dU = TotalSol.reshape(int(TotalSol.shape[0]/nvar),nvar)

        return dU

    def UpdateFreeMechanicalDoFs(self, sol, fsize, nvar):
        """Updates the geometry with iterative solutions of Newton-Raphson
            for free degrees of freedom only. Needs to be applied per time NR iteration"""

        # GET TOTAL SOLUTION
        TotalSol = np.zeros((fsize,1))
        TotalSol[self.mech_in,0] = sol

        # RE-ORDER SOLUTION COMPONENTS
        dU = TotalSol.reshape(int(TotalSol.shape[0]/nvar),nvar)

        return dU





    def LogSave(self, fem_solver, formulation, U, Eulerp, Increment):
        if fem_solver.print_incremental_log:
            dmesh = Mesh()
            dmesh.points = U
            dmesh_bounds = dmesh.Bounds
            if formulation.fields == "electro_mechanics":
                _bounds = np.zeros((2,formulation.nvar))
                _bounds[:,:formulation.ndim] = dmesh_bounds
                _bounds[:,-1] = [Eulerp.min(),Eulerp.max()]
                print("\nMinimum and maximum incremental solution values at increment {} are \n".format(Increment),_bounds)
            else:
                print("\nMinimum and maximum incremental solution values at increment {} are \n".format(Increment),dmesh_bounds)

        # SAVE INCREMENTAL SOLUTION IF ASKED FOR
        if fem_solver.save_incremental_solution:
            # FOR BIG MESHES
            if Increment % fem_solver.incremental_solution_save_frequency !=0:
                return
            from scipy.io import savemat
            filename = fem_solver.incremental_solution_filename
            if filename is not None:
                if ".mat" in filename:
                    filename = filename.split(".")[0]
                savemat(filename+"_"+str(Increment),
                    {'solution':np.hstack((U,Eulerp[:,None]))},do_compression=True)
            else:
                raise ValueError("No file name provided to save incremental solution")



import numpy as np
from .MaterialBase import Material
from Florence.Tensor import trace

class NeoHookean(Material):
    """The fundamental Neo-Hookean internal energy, described in Bonet et. al.

        W(C) = mu/2*(C:I-3)- mu*lnJ + lamb/2*(J-1)**2

    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(NeoHookean, self).__init__(mtype, ndim, **kwargs)

        self.is_transversely_isotropic = False
        self.energy_type = "internal_energy"
        self.nature = "nonlinear"
        self.fields = "mechanics"  

        if self.ndim==3:
            self.H_VoigtSize = 6
        elif self.ndim==2:
            self.H_VoigtSize = 3

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = True
        # self.has_low_level_dispatcher = False

    def KineticMeasures(self,F,ElectricFieldx=0, elem=0):
        from Florence.MaterialLibrary.LLDispatch._NeoHookean_ import KineticMeasures
        return KineticMeasures(self,F)
        

    def Hessian(self,StrainTensors,ElectricFieldx=None,elem=0,gcounter=0):
        
        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]

        mu2 = self.mu/J- self.lamb*(J-1.0)
        lamb2 = self.lamb*(2*J-1.0) 

        H_Voigt = lamb2*self.vIijIkl+mu2*self.vIikIjl

        self.H_VoigtSize = H_Voigt.shape[0]

        return H_Voigt

    def CauchyStress(self,StrainTensors,ElectricFieldx=None,elem=0,gcounter=0):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]

        mu = self.mu
        lamb = self.lamb
            
        return 1.0*mu/J*b + (lamb*(J-1.0)-mu/J)*I



    def InternalEnergy(self,StrainTensors,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        F = StrainTensors['F'][gcounter]
        C = np.dot(F.T,F)

        energy  = mu/2.*(trace(C) - 3.) - mu*np.log(J) + lamb/2.*(J-1.)**2

        return energy





#cython: profile=False
#cython: infer_types=True
#cython: boundscheck=False
#cython: wraparound=False
#cython: nonecheck=False

import numpy as np
cimport numpy as np

ctypedef double Real

cdef extern from "_MassIntegrand_.h":

    void _MassIntegrand_Filler_(Real *mass,
        const Real* bases,
        const Real* detJ,
        int ngauss,
        int noderpelem,
        int ndim,
        int nvar,
        Real rho) nogil

    void _ConstantMassIntegrand_Filler_(Real *mass,
        const Real* constant_mass_integrand,
        const Real* detJ,
        int ngauss,
        int local_capacity) nogil


def __MassIntegrand__(Real rho,
    np.ndarray[Real, ndim=2, mode='c'] bases,
    np.ndarray[Real, ndim=1] detJ,
    int ndim, int nvar):

    cdef int ngauss = detJ.shape[0]
    cdef int nodeperelem = bases.shape[0]
    cdef int local_size = nvar*nodeperelem

    cdef np.ndarray[Real, ndim=2, mode='c'] mass = np.zeros((local_size,
        local_size),dtype=np.float64)

    _MassIntegrand_Filler_(&mass[0,0],
        &bases[0,0],
        &detJ[0],
        ngauss,
        nodeperelem,
        ndim,
        nvar,
        rho)

    return mass



def __ConstantMassIntegrand__(np.ndarray[Real, ndim=3, mode='c'] constant_mass_integrand,
    np.ndarray[Real, ndim=1] detJ):

    cdef int ngauss = detJ.shape[0]
    cdef int local_size = constant_mass_integrand.shape[1]
    cdef int local_capacity = local_size*local_size

    cdef np.ndarray[Real, ndim=2, mode='c'] mass = np.zeros((local_size,
        local_size),dtype=np.float64)

    _ConstantMassIntegrand_Filler_(&mass[0,0],
        &constant_mass_integrand[0,0,0],
        &detJ[0],
        ngauss,
        local_capacity)

    return mass